import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional


if backend.is_tensorflow_backend:
    import tensorflow as tf  # Needed for all assigning achors functions


EFFICIENTDET_MODE = "efficientdet"
ANCHOR_FREE_MODE = "anchor_free"
YOLOR_MODE = "yolor"
YOLOV8_MODE = "yolov8"
NUM_ANCHORS = {ANCHOR_FREE_MODE: 1, YOLOV8_MODE: 1, YOLOR_MODE: 3, EFFICIENTDET_MODE: 9}

""" Init anchors """


def get_anchors_mode_parameters(anchors_mode, use_object_scores="auto", num_anchors="auto", anchor_scale="auto"):
    if anchors_mode == ANCHOR_FREE_MODE:
        use_object_scores = True if use_object_scores == "auto" else use_object_scores
        num_anchors = NUM_ANCHORS[anchors_mode] if num_anchors == "auto" else num_anchors
    elif anchors_mode == YOLOR_MODE:
        use_object_scores = True if use_object_scores == "auto" else use_object_scores
        num_anchors = NUM_ANCHORS[anchors_mode] if num_anchors == "auto" else num_anchors
    elif anchors_mode == YOLOV8_MODE:
        use_object_scores = False if use_object_scores == "auto" else use_object_scores
        num_anchors = NUM_ANCHORS[anchors_mode] if num_anchors == "auto" else num_anchors
    else:
        use_object_scores = False if use_object_scores == "auto" else use_object_scores
        num_anchors = NUM_ANCHORS.get(anchors_mode, NUM_ANCHORS[EFFICIENTDET_MODE]) if num_anchors == "auto" else num_anchors
        anchor_scale = 4 if anchor_scale == "auto" else anchor_scale
    return use_object_scores, num_anchors, anchor_scale


def get_feature_sizes(input_shape, pyramid_levels=[3, 7]):
    # https://github.com/google/automl/tree/master/efficientdet/utils.py#L509
    feature_sizes = [input_shape[:2]]
    for _ in range(max(pyramid_levels)):
        pre_feat_size = feature_sizes[-1]
        feature_sizes.append(((pre_feat_size[0] - 1) // 2 + 1, (pre_feat_size[1] - 1) // 2 + 1))  # ceil mode, like padding="same" downsampling
    return feature_sizes


def get_anchors(input_shape=(512, 512, 3), pyramid_levels=[3, 7], aspect_ratios=[1, 2, 0.5], num_scales=3, anchor_scale=4, grid_zero_start=False):
    """
    >>> from keras_cv_attention_models.coco import anchors_func
    >>> input_shape = [512, 128]
    >>> anchors = anchors_func.get_anchors([512, 128], pyramid_levels=[7])
    >>> anchors.draw_bboxes(anchors * [512, 128, 512, 128])

    grid_zero_start: grid starts from 0, else from strides // 2. False for efficientdet anchors, True for yolo anchors.
    """
    # base anchors
    scales = [2 ** (ii / num_scales) * anchor_scale for ii in range(num_scales)]
    aspect_ratios_tensor = np.array(aspect_ratios, dtype="float32")
    if len(aspect_ratios_tensor.shape) == 1:
        # aspect_ratios = [0.5, 1, 2]
        sqrt_ratios = np.sqrt(aspect_ratios_tensor)
        ww_ratios, hh_ratios = sqrt_ratios, 1 / sqrt_ratios
    else:
        # aspect_ratios = [(1, 1), (1.4, 0.7), (0.7, 1.4)]
        ww_ratios, hh_ratios = aspect_ratios_tensor[:, 0], aspect_ratios_tensor[:, 1]
    base_anchors_hh = np.reshape(np.expand_dims(scales, 1) * np.expand_dims(hh_ratios, 0), [-1])
    base_anchors_ww = np.reshape(np.expand_dims(scales, 1) * np.expand_dims(ww_ratios, 0), [-1])
    base_anchors_hh_half, base_anchors_ww_half = base_anchors_hh / 2, base_anchors_ww / 2
    base_anchors = np.stack([base_anchors_hh_half * -1, base_anchors_ww_half * -1, base_anchors_hh_half, base_anchors_ww_half], axis=1)
    # base_anchors = tf.gather(base_anchors, [3, 6, 0, 4, 7, 1, 5, 8, 2])  # re-order according to official generated anchors
    # For anchor_free, base_anchors = np.array([[-0.5, -0.5, 0.5, 0.5]])

    # make grid
    pyramid_levels = list(range(min(pyramid_levels), max(pyramid_levels) + 1))
    feature_sizes = get_feature_sizes(input_shape, pyramid_levels)

    all_anchors = []
    for level in pyramid_levels:
        stride_hh, stride_ww = feature_sizes[0][0] / feature_sizes[level][0], feature_sizes[0][1] / feature_sizes[level][1]
        top, left = (0, 0) if grid_zero_start else (stride_hh / 2, stride_ww / 2)
        hh_centers = np.arange(top, input_shape[0], stride_hh)
        ww_centers = np.arange(left, input_shape[1], stride_ww)
        ww_grid, hh_grid = np.meshgrid(ww_centers, hh_centers)
        grid = np.reshape(np.stack([hh_grid, ww_grid, hh_grid, ww_grid], 2), [-1, 1, 4])
        anchors = np.expand_dims(base_anchors * [stride_hh, stride_ww, stride_hh, stride_ww], 0) + grid.astype(base_anchors.dtype)
        anchors = np.reshape(anchors, [-1, 4])
        all_anchors.append(anchors)
    all_anchors = np.concatenate(all_anchors, axis=0) / [input_shape[0], input_shape[1], input_shape[0], input_shape[1]]
    # if width_first:
    #      all_anchors = tf.gather(all_anchors, [1, 0, 3, 2], axis=-1)

    return functional.convert_to_tensor(all_anchors.astype("float32"))


def get_anchor_free_anchors(input_shape=(512, 512, 3), pyramid_levels=[3, 5], grid_zero_start=True):
    return get_anchors(input_shape, pyramid_levels, aspect_ratios=[1], num_scales=1, anchor_scale=1, grid_zero_start=grid_zero_start)


def get_yolor_anchors(input_shape=(512, 512), pyramid_levels=[3, 5], offset=0.5, is_for_training=False):
    assert max(pyramid_levels) - min(pyramid_levels) < 5
    # Original yolor using width first, height first here
    if max(pyramid_levels) - min(pyramid_levels) < 3:  # [3, 5], YOLOR_CSP / YOLOR_CSPX
        anchor_ratios = np.array([[[16.0, 12], [36, 19], [28, 40]], [[75, 36], [55, 76], [146, 72]], [[110, 142], [243, 192], [401, 459]]])
        # anchor_ratios = tf.convert_to_tensor([[[13.0, 10], [30, 16], [23, 33]], [[61, 30], [45, 62], [119, 59]], [[90, 116], [198, 156], [326, 373]]])
    elif max(pyramid_levels) - min(pyramid_levels) < 4:  # [3, 6], YOLOR_*6
        anchor_ratios = np.array(
            [[[27.0, 19], [40, 44], [94, 38]], [[68, 96], [152, 86], [137, 180]], [[301, 140], [264, 303], [542, 238]], [[615, 436], [380, 739], [792, 925]]]
        )
    else:  # [3, 7] from YOLOV4_P7, using first 3 for each level
        anchor_ratios = np.array(
            [
                [[17.0, 13], [25, 22], [66, 27]],
                [[88, 57], [69, 112], [177, 69]],
                [[138, 136], [114, 287], [275, 134]],
                [[248, 268], [504, 232], [416, 445]],
                [[393, 812], [808, 477], [908, 1070]],
            ]
        )

    pyramid_levels = list(range(min(pyramid_levels), max(pyramid_levels) + 1))
    feature_sizes = get_feature_sizes(input_shape, pyramid_levels)
    # print(f"{pyramid_levels = }, {feature_sizes = }, {anchor_ratios = }")
    if is_for_training:
        # YOLOLayer https://github.com/WongKinYiu/yolor/blob/main/models/models.py#L351
        anchor_ratios = anchor_ratios[: len(pyramid_levels)] / [[[2**ii]] for ii in pyramid_levels]
        feature_sizes = np.array(feature_sizes[min(pyramid_levels) : max(pyramid_levels) + 1], "int32")
        return functional.convert_to_tensor(anchor_ratios.astype("float32")), feature_sizes

    all_anchors = []
    for level, anchor_ratio in zip(pyramid_levels, anchor_ratios):
        stride_hh, stride_ww = feature_sizes[0][0] / feature_sizes[level][0], feature_sizes[0][1] / feature_sizes[level][1]
        # hh_grid, ww_grid = tf.meshgrid(tf.range(feature_sizes[level][0]), tf.range(feature_sizes[level][1]))
        ww_grid, hh_grid = np.meshgrid(np.arange(feature_sizes[level][1]), np.arange(feature_sizes[level][0]))
        grid = np.stack([hh_grid, ww_grid], 2).astype("float32") - offset
        grid = np.reshape(grid, [-1, 1, 2])  # [1, level_feature_sizes, 2]
        cur_base_anchors = anchor_ratio[np.newaxis, :, :]  # [num_anchors, 1, 2]

        grid_nd = np.repeat(grid, cur_base_anchors.shape[1], axis=1) * [stride_hh, stride_ww]
        cur_base_anchors_nd = np.repeat(cur_base_anchors, grid.shape[0], axis=0)
        stride_nd = np.zeros_like(grid_nd) + [stride_hh, stride_ww]
        # yield grid_nd, cur_base_anchors_nd, stride_nd
        anchors = np.concatenate([grid_nd, cur_base_anchors_nd, stride_nd], axis=-1)
        all_anchors.append(np.reshape(anchors, [-1, 6]))
    all_anchors = np.concatenate(all_anchors, axis=0) / ([input_shape[0], input_shape[1]] * 3)
    return functional.convert_to_tensor(all_anchors.astype("float32"))  # [center_h, center_w, anchor_h, anchor_w, stride_h, stride_w]


def get_anchors_mode_by_anchors(input_shape, total_anchors, num_anchors="auto", pyramid_levels_min=3, num_anchors_at_each_level_cumsum=None, regression_len=4):
    if num_anchors_at_each_level_cumsum is None:
        feature_sizes = get_feature_sizes(input_shape, [pyramid_levels_min, pyramid_levels_min + 10])[pyramid_levels_min:]
        feature_sizes = np.array(feature_sizes, dtype="int32")
        num_anchors_at_each_level_cumsum = np.cumsum(np.prod(feature_sizes, axis=-1))

    if num_anchors == "auto":
        # Pick from [1, 3, 9], 1 for anchor_free, 3 for yolor, 9 for efficientdet
        picks = np.array([1, 3, 9], dtype="int32")
        max_anchors = num_anchors_at_each_level_cumsum[-1] * picks
        num_anchors = np.ceil(total_anchors / max_anchors[0]) if total_anchors > max_anchors[-1] else picks[np.argmax(total_anchors < max_anchors)]
        num_anchors = int(num_anchors)
    dd = {1: ANCHOR_FREE_MODE if regression_len == 4 else YOLOV8_MODE, 3: YOLOR_MODE, 9: EFFICIENTDET_MODE}
    return dd.get(num_anchors, EFFICIENTDET_MODE), num_anchors


def get_pyramid_levels_by_anchors(input_shape, total_anchors, num_anchors="auto", pyramid_levels_min=3):
    feature_sizes = get_feature_sizes(input_shape, [pyramid_levels_min, pyramid_levels_min + 10])[pyramid_levels_min:]
    feature_sizes = np.array(feature_sizes, dtype="int32")
    num_anchors_at_each_level_cumsum = np.cumsum(np.prod(feature_sizes, axis=-1))
    if num_anchors == "auto":
        _, num_anchors = get_anchors_mode_by_anchors(input_shape, total_anchors, num_anchors, pyramid_levels_min, num_anchors_at_each_level_cumsum)

    total_anchors = total_anchors // num_anchors
    pyramid_levels_max = pyramid_levels_min + np.argmax(num_anchors_at_each_level_cumsum > total_anchors) - 1
    return [pyramid_levels_min, int(pyramid_levels_max)]


""" Bbox functions """


def iou_nd(bboxes, anchors):
    # bboxes: [[top, left, bottom, right]], anchors: [[top, left, bottom, right]]
    anchors_nd, bboxes_nd = functional.expand_dims(anchors, 0), functional.expand_dims(bboxes, 1)
    inter_top_left = functional.maximum(anchors_nd[:, :, :2], bboxes_nd[:, :, :2])
    inter_bottom_right = functional.minimum(anchors_nd[:, :, 2:], bboxes_nd[:, :, 2:])
    inter_hw = functional.maximum(inter_bottom_right - inter_top_left, 0)
    inter_area = inter_hw[:, :, 0] * inter_hw[:, :, 1]

    bboxes_area = (bboxes[:, 2] - bboxes[:, 0]) * (bboxes[:, 3] - bboxes[:, 1])
    anchors_area = (anchors[:, 2] - anchors[:, 0]) * (anchors[:, 3] - anchors[:, 1])
    union_area = (functional.expand_dims(bboxes_area, 1) + functional.expand_dims(anchors_area, 0)) - inter_area
    return inter_area / union_area


def corners_to_center_yxhw_nd(ss):
    """input: [top, left, bottom, right], output: [center_h, center_w], [height, width]"""
    return (ss[:, :2] + ss[:, 2:]) * 0.5, ss[:, 2:] - ss[:, :2]


def center_yxhw_to_corners_nd(ss):
    """input: [center_h, center_w, height, width], output: [top, left, bottom, right]"""
    top_left = ss[:, :2] - ss[:, 2:] * 0.5
    bottom_right = top_left + ss[:, 2:]
    return functional.concat([top_left, bottom_right], axis=-1)


def _efficientdet_decode_bboxes(preds, anchors):
    preds_center, preds_hw, preds_others = functional.split(preds, [2, 2, -1], axis=-1)

    anchors_hw = anchors[:, 2:] - anchors[:, :2]
    anchors_center = (anchors[:, :2] + anchors[:, 2:]) * 0.5

    bboxes_center = preds_center * anchors_hw + anchors_center
    bboxes_hw = functional.exp(preds_hw) * anchors_hw
    return bboxes_center, bboxes_hw, preds_others


def _yolor_decode_bboxes(preds, anchors):
    preds_center, preds_hw, preds_others = functional.split(preds, [2, 2, -1], axis=-1)

    # anchors: [grid_y, grid_x, base_anchor_y, base_anchor_x, stride_y, stride_x]
    bboxes_center = preds_center * 2 * anchors[:, 4:] + anchors[:, :2]
    bboxes_hw = (preds_hw * 2) ** 2 * anchors[:, 2:4]
    return bboxes_center, bboxes_hw, preds_others


def _yolov8_decode_bboxes(preds, anchors, regression_len=64):
    preds_bbox, preds_others = functional.split(preds, [regression_len, -1], axis=-1)
    preds_bbox = functional.reshape(preds_bbox, [*preds_bbox.shape[:-1], 4, regression_len // 4])
    preds_bbox = functional.softmax(preds_bbox, axis=-1) * functional.range(preds_bbox.shape[-1], dtype="float32")
    preds_bbox = functional.reduce_sum(preds_bbox, axis=-1)
    preds_top_left, preds_bottom_right = functional.split(preds_bbox, [2, 2], axis=-1)

    anchors_hw = anchors[:, 2:] - anchors[:, :2]
    anchors_center = (anchors[:, :2] + anchors[:, 2:]) * 0.5

    bboxes_center = (preds_bottom_right - preds_top_left) / 2 * anchors_hw + anchors_center
    bboxes_hw = (preds_bottom_right + preds_top_left) * anchors_hw
    return bboxes_center, bboxes_hw, preds_others


def decode_bboxes(preds, anchors, regression_len=4, return_centers=False):
    if anchors.shape[-1] == 6:  # Currently, it's yolor / yolov7 anchors
        bboxes_center, bboxes_hw, preds_others = _yolor_decode_bboxes(preds, anchors)
    elif regression_len > 4:  # YOLOV8
        bboxes_center, bboxes_hw, preds_others = _yolov8_decode_bboxes(preds, anchors, regression_len)
    else:  # Currently, it's yolox / efficientdet anchors
        bboxes_center, bboxes_hw, preds_others = _efficientdet_decode_bboxes(preds, anchors)

    if return_centers:
        return functional.concat([bboxes_center, bboxes_hw, preds_others], axis=-1)
    else:
        preds_top_left = bboxes_center - 0.5 * bboxes_hw
        pred_bottom_right = preds_top_left + bboxes_hw
        return functional.concat([preds_top_left, pred_bottom_right, preds_others], axis=-1)


""" Assign achors """


def assign_anchor_classes_by_iou_with_bboxes(bbox_labels, anchors, ignore_threshold=0.4, overlap_threshold=0.5):
    import tensorflow as tf

    num_anchors = anchors.shape[0]
    bbox_labels = tf.gather_nd(bbox_labels, tf.where(bbox_labels[:, -1] > 0))
    bboxes, labels = bbox_labels[:, :4], bbox_labels[:, 4]

    anchor_ious = iou_nd(bboxes, anchors)  # [num_bboxes, num_anchors]
    anchor_best_iou_ids = tf.argmax(anchor_ious, axis=0)  # [num_anchors]
    # anchor_best_ious = tf.gather_nd(anchor_ious, tf.stack([anchor_best_iou_ids, tf.range(num_anchors, dtype=anchor_best_iou_ids.dtype)], axis=-1))
    anchor_best_ious = tf.reduce_max(anchor_ious, axis=0)  # This faster, [num_anchors]

    matched_idxes = tf.where(anchor_best_ious > overlap_threshold)[:, 0]
    matched_idxes = tf.unique(tf.concat([matched_idxes, tf.argmax(anchor_ious, axis=-1)], axis=0))[0]  # Ensure at leat one anchor selected for each bbox
    matched_idxes_nd = tf.expand_dims(matched_idxes, -1)
    best_match_indxes = tf.gather(anchor_best_iou_ids, matched_idxes)
    best_match_labels = tf.gather(labels, best_match_indxes)

    # Mark anchors classes, iou < ignore_threshold as 0, ignore_threshold < iou < overlap_threshold as -1
    anchor_classes = tf.where(anchor_best_ious > ignore_threshold, tf.cast(-1, bbox_labels.dtype), tf.cast(0, bbox_labels.dtype))
    # Mark matched anchors classes, iou > overlap_threshold as actual labels
    # anchor_classes = tf.where(anchor_best_ious > overlap_threshold, labels[anchor_best_iou_ids], anchor_classes)
    anchor_classes = tf.tensor_scatter_nd_update(anchor_classes, matched_idxes_nd, tf.cast(best_match_labels, bbox_labels.dtype))

    valid_anchors = tf.gather(anchors, matched_idxes)
    valid_anchors_center, valid_anchors_hw = corners_to_center_yxhw_nd(valid_anchors)
    bboxes_center, bboxes_hw = corners_to_center_yxhw_nd(bboxes)
    bboxes_centers, bboxes_hws = tf.gather(bboxes_center, best_match_indxes), tf.gather(bboxes_hw, best_match_indxes)

    encoded_anchors_center = (bboxes_centers - valid_anchors_center) / valid_anchors_hw
    encoded_anchors_hw = tf.math.log(bboxes_hws / valid_anchors_hw)
    encoded_anchors = tf.concat([encoded_anchors_center, encoded_anchors_hw], axis=-1)

    dest_boxes = tf.zeros_like(anchors)
    dest_boxes = tf.tensor_scatter_nd_update(dest_boxes, matched_idxes_nd, encoded_anchors)

    rr = tf.concat([dest_boxes, tf.expand_dims(anchor_classes, -1)], axis=-1)
    return rr


def yolor_assign_anchors(bbox_labels, anchor_ratios, feature_sizes, anchor_aspect_thresh=4.0, overlap_offset=0.5):
    """
    # Actual assigning test:
    >>> from keras_cv_attention_models import yolor, test_images
    >>> from keras_cv_attention_models.coco import anchors_func, data
    >>> mm = yolor.YOLOR_CSP()
    >>> img = test_images.dog_cat()
    >>> pred = mm(mm.preprocess_input(img))

    >>> bbs, lls, ccs = mm.decode_predictions(pred)[0]
    >>> bbox_labels = tf.concat([bbs, tf.expand_dims(tf.cast(lls, "float32"), -1)], axis=-1)
    >>> anchor_ratios, feature_sizes = anchors_func.get_yolor_anchors(mm.input_shape[1:-1], pyramid_levels=[3, 5], is_for_training=True)
    >>> assigned = anchors_func.yolor_assign_anchors(bbox_labels, anchor_ratios, feature_sizes)

    >>> # Decode and show matched
    >>> anchors = anchors_func.get_yolor_anchors(mm.input_shape[1:-1], pyramid_levels=[3, 5], is_for_training=False)
    >>> assigned, anchors = assigned[assigned[:, -1] > 0], anchors[assigned[:, -1] > 0]
    >>> decoded_centers = (assigned[:, :2] + 0.5) * anchors[:, 4:] + anchors[:, :2]
    >>> decoded_hw = assigned[:, 2:4] * anchors[:, 4:]  # assigned[:, 2:4] is multiplied with feature_size ==> assigned[:, 2:4] * strides / input_shape
    >>> decoded_corner = anchors_func.center_yxhw_to_corners_nd(tf.concat([decoded_centers, decoded_hw], axis=-1))
    >>> data.show_image_with_bboxes(test_images.dog_cat(), decoded_corner, assigned[:, -1:])
    """
    import tensorflow as tf

    bbox_labels = tf.gather_nd(bbox_labels, tf.where(bbox_labels[:, -1] > 0))
    bboxes, labels = bbox_labels[:, :4], bbox_labels[:, 4:]
    num_anchors, num_bboxes_true, num_output_channels = anchor_ratios.shape[1], tf.shape(bboxes)[0], bbox_labels.shape[-1]

    rrs = []
    # for anchor_ratio, feature_size in zip(anchor_ratios, feature_sizes):
    for id in range(feature_sizes.shape[0]):
        # build_targets https://github.dev/WongKinYiu/yolor/blob/main/utils/loss.py#L127
        anchor_ratio, feature_size = anchor_ratios[id], feature_sizes[id]
        # pick by aspect ratio
        bboxes_centers, bboxes_hws = corners_to_center_yxhw_nd(bboxes)
        bboxes_centers, bboxes_hws = bboxes_centers * feature_size, bboxes_hws * feature_size
        aspect_ratio = tf.expand_dims(bboxes_hws, 0) / tf.expand_dims(anchor_ratio, 1)
        aspect_pick = tf.reduce_max(tf.maximum(aspect_ratio, 1 / aspect_ratio), axis=-1) < anchor_aspect_thresh
        anchors_pick = tf.repeat(tf.expand_dims(tf.range(num_anchors), -1), num_bboxes_true, axis=-1)[aspect_pick]
        aspect_picked_bboxes_labels = tf.concat([bboxes_centers, bboxes_hws, labels], axis=-1)
        aspect_picked_bboxes_labels = tf.repeat(tf.expand_dims(aspect_picked_bboxes_labels, 0), num_anchors, axis=0)[aspect_pick]

        # pick by centers
        centers = aspect_picked_bboxes_labels[:, :2]
        top, left = tf.unstack(tf.logical_and(centers % 1 < overlap_offset, centers > 1), axis=-1)
        bottom, right = tf.unstack(tf.logical_and(centers % 1 > (1 - overlap_offset), centers < (feature_size - 1)), axis=-1)
        anchors_pick_all = tf.concat([anchors_pick, anchors_pick[top], anchors_pick[left], anchors_pick[bottom], anchors_pick[right]], axis=0)
        matched_top, matched_left = aspect_picked_bboxes_labels[top], aspect_picked_bboxes_labels[left]
        matched_bottom, matched_right = aspect_picked_bboxes_labels[bottom], aspect_picked_bboxes_labels[right]
        matched_bboxes_all = tf.concat([aspect_picked_bboxes_labels, matched_top, matched_left, matched_bottom, matched_right], axis=0)

        # Use matched bboxes as indexes
        matched_bboxes_idx = tf.cast(aspect_picked_bboxes_labels[:, :2], "int32")
        matched_top_idx = tf.cast(matched_top[:, :2] - [overlap_offset, 0], "int32")
        matched_left_idx = tf.cast(matched_left[:, :2] - [0, overlap_offset], "int32")
        matched_bottom_idx = tf.cast(matched_bottom[:, :2] + [overlap_offset, 0], "int32")
        matched_right_idx = tf.cast(matched_right[:, :2] + [0, overlap_offset], "int32")
        matched_bboxes_idx_all = tf.concat([matched_bboxes_idx, matched_top_idx, matched_left_idx, matched_bottom_idx, matched_right_idx], axis=0)

        # Results
        centers_true = matched_bboxes_all[:, :2] - tf.cast(matched_bboxes_idx_all, matched_bboxes_all.dtype)
        bbox_labels_true = tf.concat([centers_true, matched_bboxes_all[:, 2:]], axis=-1)
        rr = tf.zeros([feature_size[0], feature_size[1], num_anchors, num_output_channels])
        matched_bboxes_idx_all = tf.clip_by_value(matched_bboxes_idx_all, 0, tf.cast(feature_size, matched_bboxes_idx_all.dtype) - 1)
        rr = tf.tensor_scatter_nd_update(rr, tf.concat([matched_bboxes_idx_all, tf.expand_dims(anchors_pick_all, 1)], axis=-1), bbox_labels_true)
        rrs.append(tf.reshape(rr, [-1, num_output_channels]))
    return tf.concat(rrs, axis=0)


class AnchorFreeAssignMatching:
    """
    This has to be after getting model output, as picking matched anchors needs the iou value between prediction and true bboxes.

    # Basic test:
    >>> from keras_cv_attention_models.coco import anchors_func
    >>> aa = anchors_func.AnchorFreeAssignMatching([640, 640])

    >>> # Fake data
    >>> num_bboxes, num_classes, num_anchors = 32, 10, 8400
    >>> bboxes_true = tf.random.uniform([num_bboxes, 4], 0, 0.5)
    >>> bboxes_true = tf.concat([bboxes_true[:, :2], bboxes_true[:, 2:] + bboxes_true[:, :2]], axis=-1) # bottom, right > top, left
    >>> labels_true = tf.one_hot(tf.random.uniform([num_bboxes], 0, num_classes, dtype=tf.int32), num_classes)
    >>> valid_bboxes_pick = tf.cast(tf.random.uniform([num_bboxes, 1]) > 0.5, tf.float32)
    >>> bbox_labels_true = tf.concat([bboxes_true, labels_true, valid_bboxes_pick], axis=-1)
    >>> bbox_labels_pred = tf.random.uniform([num_anchors, 4 + num_classes + 1])

    >>> # Run test
    >>> bbox_labels_true_assined = aa(bbox_labels_true, bbox_labels_pred)
    >>> bboxes_true, bboxes_true_encoded, labels_true, object_true_idx_nd = tf.split(bbox_labels_true_assined, [4, 4, -1, 1], axis=-1)
    >>> object_true_idx_nd = tf.cast(object_true_idx_nd, tf.int32)
    >>> object_true_idx = object_true_idx_nd[:, 0]
    >>> object_true = tf.tensor_scatter_nd_update(tf.zeros_like(bbox_labels_pred[:, -1]), object_true_idx_nd, tf.ones_like(bboxes_true[:, -1]))
    >>> print(bboxes_true.shape, bboxes_true_encoded.shape, labels_true.shape, tf.reduce_sum(tf.cast(object_true, tf.float32)).numpy())
    >>> # (42, 4), (42, 4), (42, 10), 42.0
    >>> print(f"{object_true.shape = }, {bbox_labels_pred[object_true > 0].shape = }")
    >>> # object_true.shape = TensorShape([8400]), bbox_labels_pred[object_true > 0].shape = TensorShape([42, 15])

    # Actual assigning test:
    >>> from keras_cv_attention_models import yolox, test_images
    >>> from keras_cv_attention_models.coco import anchors_func, data
    >>> mm = yolox.YOLOXS()
    >>> img = test_images.dog_cat()
    >>> pred = mm(mm.preprocess_input(img))

    >>> aa = anchors_func.AnchorFreeAssignMatching(mm.input_shape[1:-1])
    >>> bbs, lls, ccs = mm.decode_predictions(pred)[0]
    >>> bbox_labels_true = tf.concat([bbs, tf.one_hot(lls, 80), tf.ones([bbs.shape[0], 1])], axis=-1)
    >>> bbox_labels_true_assined = aa(bbox_labels_true, pred[0])
    >>> bboxes_true, bboxes_true_encoded, labels_true, object_true_idx_nd = tf.split(bbox_labels_true_assined, [4, 4, -1, 1], axis=-1)
    >>> object_true_idx_nd = tf.cast(object_true_idx_nd, tf.int32)
    >>> object_true_idx = object_true_idx_nd[:, 0]
    >>> object_true = tf.tensor_scatter_nd_update(tf.zeros_like(pred[0, :, -1]), object_true_idx_nd, tf.ones_like(bboxes_true[:, -1]))
    >>> # Decode predictions
    >>> bbox_labels_pred_valid, anchors = pred[0][object_true > 0], aa.anchors[object_true > 0]
    >>> bboxes_preds, labels_preds, object_preds = bbox_labels_pred_valid[:, :4], bbox_labels_pred_valid[:, 4:-1], pred[0][:, -1:]
    >>> bboxes_preds_decode = anchors_func.decode_bboxes(bboxes_preds, anchors)
    >>> # Show matched predictions
    >>> data.show_image_with_bboxes(img, bboxes_preds_decode, labels_preds.numpy().argmax(-1), labels_preds.numpy().max(-1))
    >>> # Show gathered bbox ground truth
    >>> data.show_image_with_bboxes(img, bboxes_true, labels_true.numpy().argmax(-1), labels_true.numpy().max(-1))
    >>> # Show gathered encoded, bbox ground truth
    >>> bboxes_true_decode = anchors_func.decode_bboxes(bboxes_true_encoded, anchors)
    >>> data.show_image_with_bboxes(img, bboxes_true_decode, labels_true.numpy().argmax(-1), labels_true.numpy().max(-1))
    """

    def __init__(
        self,
        input_shape,
        pyramid_levels=[3, 5],
        center_radius=2.5,
        topk_ious_max=10,
        grid_zero_start=True,  # False for YOLOV8, True for YOLOX
        use_object_scores=True,  # False for YOLOV8, True for YOLOX
        regression_len=4,  # 64 fro YOLOV8, 4 for YOLOX
        with_encoded_bboxes=True,  # False for YOLOX not using l1 loss, else False
        epsilon=1e-8,
    ):
        self.center_radius, self.topk_ious_max, self.with_encoded_bboxes, self.epsilon = center_radius, topk_ious_max, with_encoded_bboxes, epsilon
        self.input_shape, self.grid_zero_start, self.use_object_scores, self.regression_len = input_shape, grid_zero_start, use_object_scores, regression_len
        # pyramid_levels = get_pyramid_levels_by_num_anchors(self.input_shape, self.num_anchors)
        self.anchors = get_anchors(self.input_shape, pyramid_levels, aspect_ratios=[1], num_scales=1, anchor_scale=1, grid_zero_start=self.grid_zero_start)

        # Anchors constant values
        self.anchors_centers = (self.anchors[:, :2] + self.anchors[:, 2:]) * 0.5
        self.anchors_hws = self.anchors[:, 2:] - self.anchors[:, :2]
        self.anchors_nd = tf.expand_dims(self.anchors, 0)  # [1, num_anchors, 4]
        self.anchors_centers_nd, self.anchors_hws_nd = tf.expand_dims(self.anchors_centers, 0), tf.expand_dims(self.anchors_hws, 0)  # [1, num_anchors, 2]
        self.centers_enlarge_nd = self.anchors_hws_nd * self.center_radius

        self.__decode_bboxes__ = self.__yolov8_decode_bboxes__ if regression_len > 4 else self.__yolox_decode_bboxes__
        self.__encode_bboxes__ = self.__yolov8_encode_bboxes__ if regression_len > 4 else self.__yolox_encode_bboxes__

    def __picking_anchors_by_center_within_bboxes__(self, bboxes_true_nd):
        # get_in_boxes_info https://github.com/Megvii-BaseDetection/YOLOX/tree/master/yolox/models/yolo_head.py#L522
        # bboxes: [[top, left, bottom, right]], anchors: [[top, left, bottom, right]]
        # anchors_centers_nd: [1, num_anchors, 2], bboxes_true_nd: [num_bboxes, 1, 4]
        # is_anchor_in_bbox: [num_bboxes, num_anchors, 2]
        is_anchor_in_bbox = tf.logical_and(bboxes_true_nd[:, :, :2] < self.anchors_centers_nd, bboxes_true_nd[:, :, 2:] > self.anchors_centers_nd)
        is_anchor_in_bbox = tf.reduce_all(is_anchor_in_bbox, axis=-1)  # All 4 points matching: [num_bboxes, num_anchors]

        bboxes_centers_nd = (bboxes_true_nd[:, :, :2] + bboxes_true_nd[:, :, 2:]) * 0.5
        is_anchor_in_center_top_left = self.anchors_centers_nd > (bboxes_centers_nd - self.centers_enlarge_nd)
        is_anchor_in_center_bottom_right = self.anchors_centers_nd < (bboxes_centers_nd + self.centers_enlarge_nd)
        is_anchor_in_center = tf.logical_and(is_anchor_in_center_top_left, is_anchor_in_center_bottom_right)
        is_anchor_in_center = tf.reduce_all(is_anchor_in_center, axis=-1)
        return is_anchor_in_bbox, is_anchor_in_center

    def __yolox_decode_bboxes__(self, bboxes_pred, anchors_centers, anchors_hws):
        bboxes_pred_center = bboxes_pred[:, :2] * anchors_hws + anchors_centers
        bboxes_pred_hw = tf.math.exp(bboxes_pred[:, 2:]) * anchors_hws
        bboxes_pred_top_left = bboxes_pred_center - 0.5 * bboxes_pred_hw
        bboxes_pred_bottom_right = bboxes_pred_top_left + bboxes_pred_hw
        return bboxes_pred_top_left, bboxes_pred_bottom_right, bboxes_pred_center, bboxes_pred_hw

    def __yolox_encode_bboxes__(self, bboxes_true, anchors_centers, anchors_hws):
        # bboxes_true_center, bboxes_true_hw = corners_to_center_yxhw_nd(bboxes_true)
        bboxes_true_hw = bboxes_true[:, 2:] - bboxes_true[:, :2]
        bboxes_true_center = (bboxes_true[:, 2:] + bboxes_true[:, :2]) / 2.0
        bboxes_true_center_encoded = (bboxes_true_center - anchors_centers) / anchors_hws
        bboxes_true_hw_encoded = tf.math.log(bboxes_true_hw / anchors_hws + self.epsilon)
        return tf.concat([bboxes_true_center_encoded, bboxes_true_hw_encoded], axis=-1)

    def __yolov8_decode_bboxes__(self, bboxes_pred, anchors_centers, anchors_hws):
        regression_len = bboxes_pred.shape[-1] // 4
        bboxes_pred = functional.reshape(bboxes_pred, [-1, *bboxes_pred.shape[1:-1], 4, regression_len])
        bboxes_pred = functional.softmax(bboxes_pred, axis=-1) * functional.range(regression_len, dtype="float32")
        bboxes_pred = functional.reduce_sum(bboxes_pred, axis=-1)
        preds_top_left, preds_bottom_right = functional.split(bboxes_pred, [2, 2], axis=-1)

        preds_top_left, preds_bottom_right = preds_top_left * anchors_hws, preds_bottom_right * anchors_hws
        bboxes_pred_hw = preds_top_left + preds_bottom_right
        bboxes_pred_top_left = anchors_centers - preds_top_left
        bboxes_pred_bottom_right = anchors_centers + preds_bottom_right
        bboxes_pred_center = (bboxes_pred_top_left + bboxes_pred_bottom_right) / 2
        return bboxes_pred_top_left, bboxes_pred_bottom_right, bboxes_pred_center, bboxes_pred_hw

    def __yolov8_encode_bboxes__(self, bboxes_true, anchors_centers, anchors_hws):
        bboxes_true_top_left, bboxes_true_bottom_right = bboxes_true[:, :2], bboxes_true[:, 2:]
        bboxes_true_top_left_encoded = (anchors_centers - bboxes_true_top_left) / anchors_hws
        bboxes_true_bottom_right_encoded = (bboxes_true_bottom_right - anchors_centers) / anchors_hws
        out = tf.concat([bboxes_true_top_left_encoded, bboxes_true_bottom_right_encoded], axis=-1)
        return tf.clip_by_value(out, 0, self.regression_len / 4 - 1.01)

    def __center_iou_nd__(self, bboxes_true_nd, bboxes_pred_top_left, bboxes_pred_bottom_right, bboxes_pred_hw):
        # bboxes_true_nd: [num_bboxes, 1, *[top, left, bottom, right]]
        bboxes_pred_top_left_nd, bboxes_pred_bottom_right_nd = tf.expand_dims(bboxes_pred_top_left, 0), tf.expand_dims(bboxes_pred_bottom_right, 0)
        inter_top_left = tf.maximum(bboxes_pred_top_left_nd, bboxes_true_nd[:, :, :2])
        inter_bottom_right = tf.minimum(bboxes_pred_bottom_right_nd, bboxes_true_nd[:, :, 2:])
        inter_hw = tf.maximum(inter_bottom_right - inter_top_left, 0)
        inter_area = inter_hw[:, :, 0] * inter_hw[:, :, 1]

        bboxes_area_nd = (bboxes_true_nd[:, :, 2] - bboxes_true_nd[:, :, 0]) * (bboxes_true_nd[:, :, 3] - bboxes_true_nd[:, :, 1])
        pred_area_nd = tf.expand_dims(bboxes_pred_hw[:, 0] * bboxes_pred_hw[:, 1], 0)
        union_area = bboxes_area_nd + pred_area_nd - inter_area
        return inter_area / union_area

    def __filter_anchors_matching_multi_bboxes__(self, topk_anchors, cost, check_cond):
        cond = tf.where(check_cond)
        conflict_costs = tf.gather(cost, cond[:, 0], axis=1)
        topk_anchors = tf.tensor_scatter_nd_update(tf.transpose(topk_anchors), cond, tf.zeros([tf.shape(cond)[0], tf.shape(topk_anchors)[0]]))
        topk_anchors = tf.transpose(topk_anchors)
        update_pos = tf.concat([tf.expand_dims(tf.argmin(conflict_costs, axis=0), 1), cond], axis=-1)  # Not considering if argmin cost is in conflict_costs
        topk_anchors = tf.tensor_scatter_nd_update(topk_anchors, update_pos, tf.ones(tf.shape(cond)[0]))
        return topk_anchors

    def __dynamic_k_matching__(self, ious, cost):
        # dynamic_k_matching https://github.com/Megvii-BaseDetection/YOLOX/tree/master/yolox/models/yolo_head.py#L607
        top_ious = tf.sort(ious, direction="DESCENDING")[:, : self.topk_ious_max]
        dynamic_ks = tf.maximum(tf.reduce_sum(top_ious, axis=-1), 1.0)  # [???] why sum up ious

        num_picked = tf.shape(cost)[-1]
        cc = tf.concat([cost, tf.expand_dims(dynamic_ks, 1)], axis=-1)
        # matching_matrix, tf.argsort default direction = "ASCENDING"
        topk_anchors = tf.map_fn(lambda xx: tf.reduce_sum(tf.one_hot(tf.argsort(xx[:-1])[: tf.cast(xx[-1], tf.int32)], num_picked), 0), cc)
        check_cond = tf.reduce_sum(topk_anchors, axis=0) > 1
        return tf.cond(
            tf.reduce_any(check_cond),
            lambda: self.__filter_anchors_matching_multi_bboxes__(topk_anchors, cost, check_cond),
            lambda: topk_anchors,
        )

    def __call__(self, bbox_labels_true, bbox_labels_pred):
        import tensorflow as tf
        from tensorflow.keras import backend as K

        # get_assignments https://github.com/Megvii-BaseDetection/YOLOX/tree/master/yolox/models/yolo_head.py#425
        bbox_labels_true = tf.gather_nd(bbox_labels_true, tf.where(bbox_labels_true[:, -1] > 0))

        bboxes_true, labels_true = bbox_labels_true[:, :4], bbox_labels_true[:, 4:-1]
        bboxes_true_nd = tf.expand_dims(bboxes_true, 1)

        # is_anchor_in_bbox, is_anchor_in_center: [num_bboxes, num_anchors]
        is_anchor_in_bbox, is_anchor_in_center = self.__picking_anchors_by_center_within_bboxes__(bboxes_true_nd)
        # [num_anchors]
        is_anchor_match_any_bbox = tf.logical_or(tf.reduce_any(is_anchor_in_bbox, axis=0), tf.reduce_any(is_anchor_in_center, axis=0))
        pick_cond = tf.where(is_anchor_match_any_bbox)[:, 0]
        # [num_bboxes, num_picked_anchors]
        is_anchor_valid = tf.logical_and(tf.gather(is_anchor_in_bbox, pick_cond, axis=-1), tf.gather(is_anchor_in_center, pick_cond, axis=-1))

        bbox_labels_pred = bbox_labels_pred[is_anchor_match_any_bbox]
        if self.use_object_scores:
            bboxes_pred, labels_pred, object_pred = tf.split(bbox_labels_pred, [self.regression_len, -1, 1], axis=-1)
            obj_labels_pred = tf.sqrt(labels_pred * object_pred)
        else:
            bboxes_pred, labels_pred = tf.split(bbox_labels_pred, [self.regression_len, -1], axis=-1)
            obj_labels_pred = tf.sqrt(labels_pred)

        # decode_bboxes
        anchors_centers, anchors_hws = self.anchors_centers[is_anchor_match_any_bbox], self.anchors_hws[is_anchor_match_any_bbox]
        bboxes_pred_top_left, bboxes_pred_bottom_right, bboxes_pred_center, bboxes_pred_hw = self.__decode_bboxes__(bboxes_pred, anchors_centers, anchors_hws)

        ious = self.__center_iou_nd__(bboxes_true_nd, bboxes_pred_top_left, bboxes_pred_bottom_right, bboxes_pred_hw)  # [num_bboxes, num_picked_anchors]
        ious_loss = -tf.math.log(ious + self.epsilon)

        cls_loss = K.binary_crossentropy(tf.expand_dims(labels_true, 1), tf.expand_dims(obj_labels_pred, 0))  # [num_bboxes, num_picked_anchors, num_classes]
        cls_loss = tf.reduce_sum(cls_loss, -1)  # [num_bboxes, num_picked_anchors]
        cost = cls_loss + 3.0 * ious_loss + 1e5 * tf.cast(tf.logical_not(is_anchor_valid), cls_loss.dtype)  # [num_bboxes, num_picked_anchors]

        # dynamic_k_matching
        bbox_matched_k_anchors = self.__dynamic_k_matching__(ious, cost)  # [num_bboxes, num_picked_anchors], contains only 0, 1
        is_anchor_iou_match_any = tf.reduce_any(bbox_matched_k_anchors > 0, axis=0)  # [num_picked_anchors]
        is_anchor_iou_match_any_idx = tf.where(is_anchor_iou_match_any)

        # TODO: is_anchor_iou_match_any_idx.shape[0] == 0
        anchor_best_matching_bbox = tf.argmax(tf.gather(bbox_matched_k_anchors, is_anchor_iou_match_any_idx[:, 0], axis=-1), 0)
        anchor_labels = tf.gather(labels_true, anchor_best_matching_bbox)  # [num_picked_anchors， num_classes]
        pred_iou_loss = tf.reduce_sum(bbox_matched_k_anchors * ious, 0)[is_anchor_iou_match_any]

        # get_losses after get_assignments. Bboxes for iou loss, [top, left, bottom, right]
        out_bboxes_true = tf.gather(bboxes_true, anchor_best_matching_bbox)
        out_labels_true = anchor_labels * tf.expand_dims(pred_iou_loss, -1)

        # object loss, [num_anchors]
        out_object_true = tf.tensor_scatter_nd_update(is_anchor_match_any_bbox, tf.where(is_anchor_match_any_bbox), is_anchor_iou_match_any)
        if tf.executing_eagerly():
            object_true_idx = tf.where(out_object_true)  # [num_picked_anchors, 1]
        else:
            with tf.xla.experimental.jit_scope(compile_ops=False):
                object_true_idx = tf.where(out_object_true)  # [num_picked_anchors, 1]

        if self.with_encoded_bboxes:
            # l1_target loss, encoded [center_top, center_left, height, width]
            anchors_centers_valid = tf.gather_nd(anchors_centers, is_anchor_iou_match_any_idx)
            anchors_hws_valid = tf.gather_nd(anchors_hws, is_anchor_iou_match_any_idx)
            out_bboxes_true_encoded = self.__encode_bboxes__(out_bboxes_true, anchors_centers_valid, anchors_hws_valid)

            # tf.stop_gradient requires returning value been a single tensor with same dtype as inputs.
            return tf.concat([out_bboxes_true, out_bboxes_true_encoded, out_labels_true, tf.cast(object_true_idx, out_bboxes_true.dtype)], axis=-1)
        else:
            return tf.concat([out_bboxes_true, out_labels_true, tf.cast(object_true_idx, out_bboxes_true.dtype)], axis=-1)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, initializers
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.attention_layers import activation_by_name, add_pre_post_process

BATCH_NORM_EPSILON = 1e-5
PRETRAINED_DICT = {
    "gmlp_s16": {"imagenet": "cc2d83bc0a7edd257aa6cd58397887e9"},
}


def layer_norm(inputs, name=None):
    """Typical LayerNormalization with epsilon=1e-5"""
    norm_axis = -1 if backend.image_data_format() == "channels_last" else 1
    return layers.LayerNormalization(axis=norm_axis, epsilon=BATCH_NORM_EPSILON, name=name)(inputs)


def spatial_gating_block(inputs, name=None):
    uu, vv = functional.split(inputs, 2, axis=-1 if backend.image_data_format() == "channels_last" else 1)
    # print(f">>>> {uu.shape = }, {vv.shape = }")
    vv = layer_norm(vv, name=name and name + "vv_ln")
    vv = layers.Permute((2, 1), name=name and name + "permute_1")(vv) if backend.image_data_format() == "channels_last" else vv
    ww_init = initializers.truncated_normal(stddev=1e-6)
    vv = layers.Dense(vv.shape[-1], kernel_initializer=ww_init, bias_initializer="ones", name=name and name + "vv_dense")(vv)
    vv = layers.Permute((2, 1), name=name and name + "permute_2")(vv) if backend.image_data_format() == "channels_last" else vv
    # print(f">>>> {uu.shape = }, {vv.shape = }")
    gated_out = layers.Multiply()([uu, vv])
    return gated_out


def res_gated_mlp_block(inputs, channels_mlp_dim, drop_rate=0, activation="gelu", name=None):
    input_channel = inputs.shape[-1] if backend.image_data_format() == "channels_last" else inputs.shape[1]
    nn = layer_norm(inputs, name=name + "pre_ln")
    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute((2, 1), name=name and name + "pre_dense_permute_1")(nn)
    nn = layers.Dense(channels_mlp_dim, name=name + "pre_dense")(nn)
    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute((2, 1), name=name and name + "pre_dense_permute_2")(nn)
    nn = activation_by_name(nn, activation, name=name and name + activation)
    # Drop

    # print(f">>>> {inputs.shape = }, {nn.shape = }")
    # SpatialGatingUnit
    nn = spatial_gating_block(nn, name=name)
    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute((2, 1), name=name and name + "gated_dense_permute_1")(nn)
    nn = layers.Dense(input_channel, name=name + "gated_dense")(nn)
    nn = nn if backend.image_data_format() == "channels_last" else layers.Permute((2, 1), name=name and name + "gated_dense_permute_2")(nn)
    # Drop

    # Drop path
    if drop_rate > 0:
        nn = layers.Dropout(drop_rate, noise_shape=(None, 1, 1), name=name + "drop")(nn)
    return layers.Add(name=name + "output")([nn, inputs])


def GMLP(
    num_blocks,
    patch_size,
    stem_width,
    channels_mlp_dim,
    input_shape=(224, 224, 3),
    num_classes=0,
    activation="gelu",
    sam_rho=0,
    dropout=0,
    drop_connect_rate=0,
    classifier_activation="softmax",
    pretrained="imagenet",
    model_name="gmlp",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    nn = layers.Conv2D(stem_width, kernel_size=patch_size, strides=patch_size, padding="valid", name="stem")(inputs)
    new_shape = [nn.shape[1] * nn.shape[2], stem_width] if backend.image_data_format() == "channels_last" else [stem_width, nn.shape[2] * nn.shape[3]]
    nn = layers.Reshape(new_shape)(nn)

    drop_connect_s, drop_connect_e = drop_connect_rate if isinstance(drop_connect_rate, (list, tuple)) else [drop_connect_rate, drop_connect_rate]
    for ii in range(num_blocks):
        name = "{}_{}_".format("gmlp", str(ii + 1))
        block_drop_rate = drop_connect_s + (drop_connect_e - drop_connect_s) * ii / num_blocks
        nn = res_gated_mlp_block(nn, channels_mlp_dim=channels_mlp_dim, drop_rate=block_drop_rate, activation=activation, name=name)
    nn = layer_norm(nn, name="pre_head_norm")

    if num_classes > 0:
        # nn = tf.reduce_mean(nn, axis=1)
        nn = layers.GlobalAveragePooling1D()(nn)
        if dropout > 0 and dropout < 1:
            nn = layers.Dropout(dropout)(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    if sam_rho != 0:
        from keras_cv_attention_models.model_surgery import SAMModel

        model = SAMModel(inputs, nn, name=model_name)
    else:
        model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="tf")
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="mlp_family", pretrained=pretrained)
    return model


BLOCK_CONFIGS = {
    "tiny16": {
        "num_blocks": 30,
        "patch_size": 16,
        "stem_width": 128,
        "channels_mlp_dim": 128 * 6,
    },
    "s16": {
        "num_blocks": 30,
        "patch_size": 16,
        "stem_width": 256,
        "channels_mlp_dim": 256 * 6,
    },
    "b16": {
        "num_blocks": 30,
        "patch_size": 16,
        "stem_width": 512,
        "channels_mlp_dim": 512 * 6,
    },
}


@register_model
def GMLPTiny16(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return GMLP(**BLOCK_CONFIGS["tiny16"], **locals(), model_name="gmlp_tiny16", **kwargs)


@register_model
def GMLPS16(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return GMLP(**BLOCK_CONFIGS["s16"], **locals(), model_name="gmlp_s16", **kwargs)


@register_model
def GMLPB16(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return GMLP(**BLOCK_CONFIGS["b16"], **locals(), model_name="gmlp_b16", **kwargs)



from keras_cv_attention_models.mobilenetv3_family.mobilenetv3 import MobileNetV3
from keras_cv_attention_models.models import register_model


def LCNet(
    num_blocks=[1, 2, 2, 1, 5, 2],
    out_channels=[32, 64, 128, 256, 256, 512],
    expands=1,
    kernel_sizes=[3, 3, 3, 3, 5, 5],
    strides=[1, 2, 2, 2, 1, 2],
    activations="hard_swish",
    disable_shortcut=True,
    use_blocks_output_activation=True,
    se_ratios=[0, 0, 0, 0, 0, 0.25],
    output_num_features=1280,
    use_additional_output_conv=False,
    model_name="lcnet",
    **kwargs,
):
    kwargs.pop("kwargs", None)
    return MobileNetV3(**locals(), **kwargs)


@register_model
def LCNet050(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return LCNet(**locals(), width_ratio=0.5, model_name="lcnet_050", **kwargs)


@register_model
def LCNet075(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return LCNet(**locals(), width_ratio=0.75, model_name="lcnet_075", **kwargs)


@register_model
def LCNet100(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return LCNet(**locals(), model_name="lcnet_100", **kwargs)


@register_model
def LCNet150(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    use_output_feature_bias = False
    return LCNet(**locals(), width_ratio=1.5, model_name="lcnet_150", **kwargs)


@register_model
def LCNet200(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    use_output_feature_bias = False
    return LCNet(**locals(), width_ratio=2.0, model_name="lcnet_200", **kwargs)


@register_model
def LCNet250(input_shape=(224, 224, 3), num_classes=1000, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    use_output_feature_bias = False
    return LCNet(**locals(), width_ratio=2.5, model_name="lcnet_250", **kwargs)



from keras_cv_attention_models.ghostnet.ghostnet_v2 import GhostNetV2, GhostNetV2_100, GhostNetV2_130, GhostNetV2_160
from keras_cv_attention_models.ghostnet.ghostnet import GhostNet, GhostNet_050, GhostNet_100, GhostNet_130

__v2_head_doc__ = """
Keras implementation of [Gitee mindspore/models/ghostnetv2](https://gitee.com/mindspore/models/tree/master/research/cv/ghostnetv2).
Paper [PDF GhostNetV2: Enhance Cheap Operation with Long-Range Attention](https://openreview.net/pdf/6db544c65bbd0fa7d7349508454a433c112470e2.pdf).
"""

__tail_doc__ = """  kernel_sizes: kernel_size for each stack.
  first_ghost_channels: num channels for first ghost module in each stack.
  out_channels: output channels for each stack.
  se_ratios: se_ratio for each stack.
  strides: stride for each stack.
  stem_width: output dimension for stem block.
  stem_strides: strides for stem `Conv2D`, default `2`.
  num_ghost_module_v1_stacks: num of `ghost_module` stcks on the head, others are `ghost_module_multiply`.
      - for `GhostNet` v1 way, default `-1` for all using `ghost_module`.
      - for `GhostNetV2` way, default `2` for only using `ghost_module` in the first 2 stacks.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default "relu".
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: One of `[None, "imagenet", "ssld"]`. "ssld" if for `GhostNet_130`.
  **kwargs: other parameters if available.

Returns:
  A `keras.Model` instance.
"""

GhostNetV2.__doc__ = __v2_head_doc__ + """
Args:
  width_mul: expansion ratio of `fist_ghost_channels` and `out_channels` in each block.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model             | Params | FLOPs  | Input | Top1 Acc |
  | ----------------- | ------ | ------ | ----- | -------- |
  | GhostNetV2_100    | 6.12M  | 168.5M | 224   | 74.41    |
  | GhostNetV2 (1.0x) | 6.12M  | 168.5M | 224   | 75.3     |
  | GhostNetV2 (1.3x) | 8.96M  | 271.1M | 224   | 76.9     |
  | GhostNetV2 (1.6x) | 12.39M | 400.9M | 224   | 77.8     |
"""

GhostNetV2_100.__doc__ = __v2_head_doc__ + """
Args:
""" + __tail_doc__

GhostNetV2_130.__doc__ = GhostNetV2_100.__doc__
GhostNetV2_160.__doc__ = GhostNetV2_100.__doc__

__v1_head_doc__ = """
Keras implementation of [Github huawei-noah/ghostnet_pytorch](https://github.com/huawei-noah/Efficient-AI-Backbones/tree/master/ghostnet_pytorch).
Paper [PDF 1911.11907 GhostNet: More Features from Cheap Operations](https://arxiv.org/pdf/1911.11907.pdf).
"""

GhostNet.__doc__ = __v1_head_doc__ + """
Args:
  width_mul: expansion ratio of `fist_ghost_channels` and `out_channels` in each block.
  stem_width: output dimension for stem block.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model        | Params | FLOPs  | Input | Top1 Acc |
  | ------------ | ------ | ------ | ----- | -------- |
  | GhostNet_050 | 2.59M  | 42.6M  | 224   | 66.88    |
  | GhostNet_100 | 5.18M  | 141.7M | 224   | 74.16    |
  | GhostNet_130 | 7.36M  | 227.7M | 224   | 75.79    |
  | - ssld       | 7.36M  | 227.7M | 224   | 79.38    |
"""

GhostNet_050.__doc__ = __v1_head_doc__ + """
Args:
""" + __tail_doc__

GhostNet_100.__doc__ = GhostNet_050.__doc__
GhostNet_130.__doc__ = GhostNet_050.__doc__



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    drop_block,
    group_norm,
    # mlp_block, # cannot import name 'mlp_block' due to circular import
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "wavemlp_t": {"imagenet": "c8fe3c22c129180c5cff7b734ede831c"},
    "wavemlp_s": {"imagenet": "b0375b030d1c3012232286cff76d301d"},
    "wavemlp_m": {"imagenet": "d2364156144ab259069026b417ca21da"},
}


def mlp_block(inputs, hidden_dim, out_channel, drop_rate=0, activation="gelu", name=""):
    nn = layers.Conv2D(hidden_dim, kernel_size=1, use_bias=True, name=name and name + "Conv_0")(inputs)
    nn = activation_by_name(nn, activation, name=name and name + activation)
    nn = layers.Dropout(drop_rate) if drop_rate > 0 else nn
    nn = layers.Conv2D(out_channel, kernel_size=1, use_bias=True, name=name and name + "Conv_1")(nn)
    nn = layers.Dropout(drop_rate) if drop_rate > 0 else nn
    return nn


def phase_aware_token_mixing(inputs, out_channel=-1, qkv_bias=False, output_dropout=0, activation="gelu", name=None):
    input_channel = inputs.shape[-1] if backend.image_data_format() == "channels_last" else inputs.shape[1]
    out_channel = out_channel if out_channel > 0 else input_channel

    theta_h = conv2d_no_bias(inputs, out_channel, kernel_size=1, use_bias=True, name=name and name + "theta_h_")
    theta_h = batchnorm_with_activation(theta_h, activation="relu", name=name and name + "theta_h_")  # Fixed as relu [ ??? ]
    height = conv2d_no_bias(inputs, out_channel, kernel_size=1, use_bias=qkv_bias, name=name and name + "height_")
    # height = layers.Concatenate(axis=-1)([height * functional.cos(theta_h), height * functional.sin(theta_h)])
    height_cos = layers.Multiply()([height, functional.cos(theta_h)])
    height_sin = layers.Multiply()([height, functional.sin(theta_h)])
    height = layers.Concatenate(axis=-1 if backend.image_data_format() == "channels_last" else 1)([height_cos, height_sin])
    height = conv2d_no_bias(height, out_channel, kernel_size=(1, 7), padding="same", groups=out_channel, use_bias=False, name=name and name + "height_down_")

    theta_w = conv2d_no_bias(inputs, out_channel, kernel_size=1, use_bias=True, name=name and name + "theta_w_")
    theta_w = batchnorm_with_activation(theta_w, activation="relu", name=name and name + "theta_w_")  # Fixed as relu [ ??? ]
    width = conv2d_no_bias(inputs, out_channel, kernel_size=1, use_bias=qkv_bias, name=name and name + "width_")
    # width = layers.Concatenate(axis=-1)([width * functional.cos(theta_w), width * functional.sin(theta_w)])
    width_cos = layers.Multiply()([width, functional.cos(theta_w)])
    width_sin = layers.Multiply()([width, functional.sin(theta_w)])
    width = layers.Concatenate(axis=-1 if backend.image_data_format() == "channels_last" else 1)([width_cos, width_sin])
    width = conv2d_no_bias(width, out_channel, kernel_size=(7, 1), padding="same", groups=out_channel, use_bias=False, name=name and name + "width_down_")

    channel = conv2d_no_bias(inputs, out_channel, kernel_size=1, use_bias=qkv_bias, name=name and name + "channel_")

    # print(f"{height.shape = }, {width.shape = }, {channel.shape = }, {out_channel = }")
    nn = layers.Add(name=name and name + "combine")([height, width, channel])
    nn = layers.GlobalAveragePooling2D(keepdims=True)(nn)
    nn = mlp_block(nn, out_channel // 4, out_channel=out_channel * 3, activation=activation, name=name and name + "reweight_")
    nn = layers.Reshape([1, 1, out_channel, 3] if backend.image_data_format() == "channels_last" else [out_channel, 1, 1, 3])(nn)
    nn = layers.Softmax(axis=-1, name=name and name + "attention_scores")(nn)
    attn_height, attn_width, attn_channel = functional.unstack(nn, axis=-1)
    # attn = layers.Add()([height * attn_height, width * attn_width, channel * attn_channel])
    attn_height = layers.Multiply()([height, attn_height])
    attn_width = layers.Multiply()([width, attn_width])
    attn_channel = layers.Multiply()([channel, attn_channel])
    # print(f"{attn_height.shape = }, {attn_width.shape = }, {attn_channel.shape = }")
    attn = layers.Add()([attn_height, attn_width, attn_channel])

    out = conv2d_no_bias(attn, out_channel, kernel_size=1, use_bias=True, name=name and name + "out_")
    out = layers.Dropout(output_dropout, name=name and name + "out_drop")(out) if output_dropout > 0 else out
    return out


def wave_block(inputs, qkv_bias=False, mlp_ratio=4, use_group_norm=False, drop_rate=0, activation="gelu", name=""):
    input_channel = inputs.shape[-1] if backend.image_data_format() == "channels_last" else inputs.shape[1]

    attn = group_norm(inputs, groups=1, name=name + "attn_") if use_group_norm else batchnorm_with_activation(inputs, activation=None, name=name + "attn_")
    attn = phase_aware_token_mixing(attn, out_channel=input_channel, qkv_bias=qkv_bias, activation=activation, name=name + "attn_")
    attn = drop_block(attn, drop_rate=drop_rate, name=name + "attn_")
    attn_out = layers.Add(name=name + "attn_out")([inputs, attn])

    mlp = group_norm(attn_out, groups=1, name=name + "mlp_") if use_group_norm else batchnorm_with_activation(attn_out, activation=None, name=name + "mlp_")
    mlp = mlp_block(mlp, int(input_channel * mlp_ratio), out_channel=input_channel, activation=activation, name=name + "mlp_")
    mlp = drop_block(mlp, name=name + "mlp_")
    mlp_out = layers.Add(name=name + "mlp_out")([attn_out, mlp])
    return mlp_out


def WaveMLP(
    num_blocks=[2, 2, 4, 2],
    out_channels=[64, 128, 320, 512],
    stem_width=-1,
    mlp_ratios=[4, 4, 4, 4],
    use_downsample_norm=True,
    use_group_norm=False,
    qkv_bias=False,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    sam_rho=0,
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="wavemlp",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    nn = layers.ZeroPadding2D(padding=2, name="stem_pad")(inputs)
    nn = conv2d_no_bias(nn, stem_width, kernel_size=7, strides=4, padding="valid", use_bias=True, name="stem_")
    if use_downsample_norm:
        nn = group_norm(nn, groups=1, name="stem_") if use_group_norm else batchnorm_with_activation(nn, activation=None, name="stem_")

    """ stage [1, 2, 3, 4] """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel, mlp_ratio) in enumerate(zip(num_blocks, out_channels, mlp_ratios)):
        stage_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            nn = conv2d_no_bias(nn, out_channel, kernel_size=3, strides=2, padding="same", use_bias=True, name=stage_name + "down_sample_")
            if use_downsample_norm:
                norm_name = stage_name + "down_sample_"
                nn = group_norm(nn, groups=1, name=norm_name) if use_group_norm else batchnorm_with_activation(nn, activation=None, name=norm_name)
        for block_id in range(num_block):
            name = stage_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            global_block_id += 1
            nn = wave_block(nn, qkv_bias, mlp_ratio, use_group_norm, block_drop_rate, activation=activation, name=name)

    if num_classes > 0:
        nn = group_norm(nn, groups=1, name="output_") if use_group_norm else batchnorm_with_activation(nn, activation=None, name="output_")
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    if sam_rho != 0:
        from keras_cv_attention_models.model_surgery import SAMModel

        model = SAMModel(inputs, nn, name=model_name)
    else:
        model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "mlp_family", pretrained)
    return model


@register_model
def WaveMLP_T(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 4, 2]
    out_channels = [64, 128, 320, 512]
    return WaveMLP(**locals(), model_name="wavemlp_t", **kwargs)


@register_model
def WaveMLP_S(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 3, 10, 3]
    out_channels = [64, 128, 320, 512]
    use_group_norm = True
    return WaveMLP(**locals(), model_name="wavemlp_s", **kwargs)


@register_model
def WaveMLP_M(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 18, 3]
    out_channels = [64, 128, 320, 512]
    mlp_ratios = [8, 8, 4, 4]
    use_group_norm = True
    use_downsample_norm = False
    return WaveMLP(**locals(), model_name="wavemlp_m", **kwargs)


@register_model
def WaveMLP_B(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained=None, **kwargs):
    num_blocks = [2, 2, 18, 2]
    out_channels = [96, 192, 384, 768]
    use_group_norm = True
    use_downsample_norm = False
    return WaveMLP(**locals(), model_name="wavemlp_b", **kwargs)



import torch
import numpy as np
from tqdm.auto import tqdm
from torch import nn
from contextlib import nullcontext
from keras_cv_attention_models import backend
from keras_cv_attention_models.pytorch_backend import layers, callbacks, metrics


class _Trainer_(object):
    def init_metrics(self, cur_metrics=None):
        if cur_metrics is None:
            metrics_names, cur_metrics = [], []
        elif isinstance(cur_metrics, str):
            metrics_names, cur_metrics = [cur_metrics], [cur_metrics]
        elif isinstance(cur_metrics, (list, tuple)):
            metrics_names, cur_metrics = [ii if isinstance(ii, str) else ii.name for ii in cur_metrics], list(cur_metrics)
        elif isinstance(cur_metrics, dict):
            assert self.output_names is not None, "self.output_names cannot be None when using dict metrics, provide output_names or use list metrics"
            assert len(self.output_names) == len(cur_metrics), "self.output_names={} not matching with metrics={}".format(self.output_names, metrics)
            metrics_names, cur_metrics = self.output_names, [cur_metrics[ii] for ii in self.output_names]
        else:
            metrics_names, cur_metrics = [cur_metrics.name], [cur_metrics]
        cur_metrics = [metrics.BUILDIN_METRICS[ii]() if isinstance(ii, str) else ii for ii in cur_metrics]
        return metrics_names, cur_metrics

    @property
    def compute_dtype(self):
        try:
            return next(self.parameters()).dtype
        except StopIteration:
            return torch.get_default_dtype()

    def train_compile(self, optimizer="RMSprop", loss=None, metrics=None, loss_weights=None, grad_accumulate=1, grad_max_norm=-1, **kwargs):
        # works like kers `model.compile`, but `compile` is took by `nn.Module`, rename as `train_compile`
        self.optimizer = getattr(torch.optim, optimizer)(self.parameters()) if isinstance(optimizer, str) else optimizer
        self.compiled_loss = self.loss = (lambda y_true, y_pred: torch.functional.F.cross_entropy(y_pred, y_true)) if loss is None else loss
        self.loss_weights, self.grad_accumulate, self.grad_max_norm = loss_weights or 1.0, grad_accumulate, grad_max_norm
        self.metrics_names, self.metrics = self.init_metrics(metrics)
        self.eval_metrics = [metric.name for metric in self.metrics if metric.eval_only]  # Mark metric like acc5 for eval only

        device = next(self.parameters()).device
        device_type = device.type
        if device_type == "cpu":
            scaler = torch.cuda.amp.GradScaler(enabled=False)
            global_context = nullcontext()
        else:
            scaler = torch.cuda.amp.GradScaler(enabled=True)
            global_context = torch.autocast(device_type=device_type, dtype=torch.float16)
        self.device, self.device_type, self.scaler, self.global_context = device, device_type, scaler, global_context

    def _convert_data_(self, data):
        if isinstance(data, np.ndarray):
            data = torch.from_numpy(data)
        if self.device_type == "cuda":
            # pin arrays x,y, which allows us to move them to GPU asynchronously (non_blocking=True)
            data = data.pin_memory().to(self.device, non_blocking=True)
        return data

    def train_step(self, xx, yy):
        # Split out for being able to overwrite
        with self.global_context:
            out = self(xx)
            loss = self.loss(yy, out) * self.loss_weights
        return out, loss

    def fit(self, x=None, y=None, batch_size=32, epochs=1, callbacks=None, validation_data=None, initial_epoch=0, validation_batch_size=None, **kwargs):
        callbacks = callbacks or []
        [ii.set_model(self) for ii in callbacks if ii.model is None]
        self.history = {"loss": []}
        validation_batch_size = validation_batch_size or batch_size
        self.batch_size, self.callbacks, self.validation_batch_size = batch_size, callbacks, validation_batch_size

        bar_format = "{n_fmt}/{total_fmt} [{bar:30}] - ETA: {elapsed}<{remaining} {rate_fmt}{postfix}{desc}"
        for epoch in range(initial_epoch, epochs):
            print("Epoch {}/{}".format(epoch + 1, epochs))
            epoch_logs = {}  # Can be used as global value between different callbacks
            [ii.on_epoch_begin(epoch, epoch_logs) for ii in callbacks]
            [ii.reset_state() for ii in self.metrics]

            self.train()
            self.optimizer.zero_grad()

            avg_loss, accumulate_passed_batches = 0.0, 0
            train_dataset, total = self._dataset_gen_(x, y, batch_size=batch_size)
            process_bar = tqdm(enumerate(train_dataset), total=total, bar_format=bar_format, ascii=".>>=")
            for batch, (xx, yy) in process_bar:
                batch_logs = {}  # Can be used as global value between different callbacks
                [ii.on_train_batch_begin(batch, batch_logs) for ii in callbacks]

                xx = [self._convert_data_(ii) for ii in xx] if isinstance(xx, (list, tuple)) else self._convert_data_(xx)
                yy = [self._convert_data_(ii) for ii in yy] if isinstance(yy, (list, tuple)) else self._convert_data_(yy)
                out, loss = self.train_step(xx, yy)
                self.scaler.scale(loss).backward()

                accumulate_passed_batches += 1
                if accumulate_passed_batches >= self.grad_accumulate:
                    self.scaler.unscale_(self.optimizer)
                    if self.grad_max_norm > 0:
                        torch.nn.utils.clip_grad_norm_(self.parameters(), max_norm=self.grad_max_norm)  # clip gradients
                    self.scaler.step(self.optimizer)  # self.optimizer.step()
                    self.scaler.update()
                    self.optimizer.zero_grad()
                    accumulate_passed_batches = 0
                # print(">>>> Epoch {}, batch: {}, loss: {:.4f}".format(epoch, batch, loss.item()))
                avg_loss += loss
                process_bar.desc = " - loss: {:.4f}".format(avg_loss / (batch + 1))  # process_bar.set_description automatically add a : on the tail

                # out = out.detach().cpu()
                if isinstance(yy, (list, tuple)):
                    [metric.update_state(cur_yy, cur_out) for cur_out, cur_yy, metric in zip(out, yy, self.metrics) if metric.name not in self.eval_metrics]
                else:
                    [metric.update_state(yy, out) for metric in self.metrics if metric.name not in self.eval_metrics]
                metrics_results = {ii: metric.result().item() for ii, metric in zip(self.metrics_names, self.metrics) if metric.name not in self.eval_metrics}
                process_bar.desc += "".join([" - {}: {:.4f}".format(name, metric) for name, metric in metrics_results.items()])

                batch_logs["loss"] = loss.item()
                batch_logs.update(metrics_results)
                [ii.on_train_batch_end(batch, logs=batch_logs) for ii in callbacks]

                """ Eval process, put inside for loop for a better display of process bar """
                if batch == total - 1 and validation_data is not None:
                    val_loss, val_metrics_results = self.evaluate(validation_data, batch_size=validation_batch_size, callbacks=callbacks)

                    self.history.setdefault("val_loss", []).append(val_loss.item())
                    for name, val_metric_result in val_metrics_results.items():
                        self.history.setdefault(name, []).append(val_metric_result.item() if hasattr(val_metric_result, "item") else val_metric_result)

                    process_bar.desc += " - val_loss: {:.4f}".format(val_loss)
                    process_bar.desc += "".join([" - {}: {:.4f}".format(name, metric) for name, metric in val_metrics_results.items()])
                    # process_bar.display()
                process_bar.refresh()

            loss = avg_loss / (batch + 1)
            self.history["loss"].append(loss.item() if hasattr(loss, "item") else loss)
            for name, metric_result in metrics_results.items():
                self.history.setdefault(name, []).append(metric_result)
            epoch_logs = {kk: vv[-1] for kk, vv in self.history.items()}
            with self.global_context, torch.no_grad():
                [ii.on_epoch_end(epoch, epoch_logs) for ii in callbacks]
            print()

    def evaluate(self, x=None, y=None, batch_size=None, verbose="auto", callbacks=None, **kwargs):
        callbacks = callbacks or []
        epoch_logs = {}  # Can be used as global value between different callbacks
        [ii.set_model(self) for ii in callbacks if ii.model is None]
        [ii.on_test_begin(epoch_logs) for ii in callbacks]
        [ii.reset_state() for ii in self.metrics]

        self.eval()
        avg_loss, accumulate_passed_batches = 0.0, 0
        val_dataset, total = self._dataset_gen_(x, y, batch_size=batch_size)
        for batch, (xx, yy) in enumerate(val_dataset):
            batch_logs = {}  # Can be used as global value between different callbacks
            [ii.on_test_batch_begin(batch, batch_logs) for ii in callbacks]

            xx = [self._convert_data_(ii) for ii in xx] if isinstance(xx, (list, tuple)) else self._convert_data_(xx)
            yy = [self._convert_data_(ii) for ii in yy] if isinstance(yy, (list, tuple)) else self._convert_data_(yy)
            with torch.no_grad():
                out, loss = self.train_step(xx, yy)
            avg_loss += loss

            # out = out.detach().cpu()
            if isinstance(yy, (list, tuple)):
                [metric.update_state(cur_yy, cur_out) for cur_out, cur_yy, metric in zip(out, yy, self.metrics)]
            else:
                [ii.update_state(yy, out) for ii in self.metrics]
            batch_logs["val_loss"] = loss.item()
            batch_logs.update({name: metric.result().item() for name, metric in zip(self.metrics_names, self.metrics)})
            [ii.on_test_batch_end(batch, batch_logs) for ii in callbacks]

        val_loss = avg_loss / (batch + 1)
        metrics_results = {}
        for name, metric in zip(self.metrics_names, self.metrics):
            metrics_results["val_" + name] = metric.result().item()
        epoch_logs["val_loss"] = val_loss.item()
        epoch_logs.update(metrics_results)
        [ii.on_test_end(epoch_logs) for ii in callbacks]
        return val_loss, metrics_results

    def _dataset_gen_(self, x=None, y=None, batch_size=32):
        if isinstance(x, (list, tuple)) and len(x) == 2 and y is None:
            xx, yy = x[0], x[1]
        else:
            xx, yy = x, y

        if hasattr(xx, "element_spec"):  # TF datsets
            data_shape = xx.element_spec[0].shape
            if self.input_shape is not None and data_shape[-1] == self.input_shape[1]:
                perm = [0, len(data_shape) - 1] + list(range(1, len(data_shape) - 1))  # [0, 3, 1, 2]
                dataset = ((ii.transpose(perm), jj) for ii, jj in xx.as_numpy_iterator())
            else:
                dataset = xx.as_numpy_iterator()
            total = len(xx)
        elif isinstance(xx, np.ndarray) or isinstance(xx, torch.Tensor):
            assert yy is not None
            num_batches = xx.shape[0] if batch_size is None else int(np.ceil(xx.shape[0] / batch_size))

            def _convert_tensor(data, id):
                cur = data[id * batch_size : (id + 1) * batch_size] if batch_size is not None else data[id]
                cur = torch.from_numpy(cur) if isinstance(cur, np.ndarray) else cur
                cur = cur.float() if cur.dtype == torch.float64 else cur
                cur = cur.long() if cur.dtype == torch.int32 else cur
                return cur

            dataset = ((_convert_tensor(xx, id), _convert_tensor(yy, id)) for id in range(num_batches))
            total = num_batches
        else:  # generator or torch.utils.data.DataLoader
            dataset = xx
            total = len(xx)
        return dataset, total


class _Exporter_(object):
    def _create_fake_input_data_(self, input_shape=None, batch_size=1):
        input_shapes = [ii.shape for ii in self.inputs]
        input_dtypes = [ii.dtype for ii in self.inputs]

        input_shape = self.input_shape if input_shape is None else input_shape
        input_shapes = input_shape if isinstance(input_shape[0], (list, tuple)) else [input_shape]  # Convert to list of input_shpae
        assert len(input_shapes) == len(self.inputs), "input_shape={} length not matching self.inputs={}".format(input_shape, self.inputs)

        input_datas = []
        for input_shape, model_input in zip(input_shapes, self.inputs):
            input_shape = list(input_shape).copy()
            if len(input_shape) == len(model_input.shape) - 1:
                input_shape = [batch_size] + input_shape
            assert len(input_shape) == len(model_input.shape), "input_shape={} rank not match with input={}".format(input_shape, model_input.shape)

            if input_shape[0] is None or input_shape[0] == -1:
                input_shape[0] = batch_size
            if None in input_shape or -1 in input_shape:
                print("[WARNING] dynamic shape value in input_shape={}, set to 32".format(input_shape))
                input_shape = [32 if ii is None or ii == -1 else ii for ii in input_shape]

            dtype = model_input.dtype or torch.get_default_dtype()
            dtype = getattr(torch, dtype) if isinstance(dtype, str) else dtype
            input_datas.append(torch.ones(input_shape, dtype=dtype))
        print(">>>> input_shape: {}, dtype: {}".format([ii.shape for ii in input_datas], [ii.dtype for ii in input_datas]))
        return input_datas

    def summary(self, input_shape=None, **kwargs):
        from torchinfo import summary

        input_datas = self._create_fake_input_data_(input_shape)
        print(summary(self, input_data=input_datas if len(self.inputs) == 1 else [input_datas], **kwargs))

    def export_onnx(self, filepath=None, input_shape=None, batch_size=1, simplify=False, input_names=None, output_names=None, **kwargs):
        input_datas = self._create_fake_input_data_(input_shape, batch_size=batch_size)

        dynamic_axes = kwargs.pop("dynamic_axes", None)
        input_names = input_names or self.input_names
        output_names = output_names or self.output_names
        if dynamic_axes is None and (batch_size is None or batch_size == -1):
            print("Set dynamic batch size")
            dynamic_axes = {ii: {0: "-1"} for ii in input_names}
            dynamic_axes.update({ii: {0: "-1"} for ii in output_names})

        filepath = (self.name + ".onnx") if filepath is None else (filepath if filepath.endswith(".onnx") else (filepath + ".onnx"))
        torch.onnx.export(self, input_datas, filepath, input_names=input_names, output_names=output_names, dynamic_axes=dynamic_axes, **kwargs)
        print("Exported onnx:", filepath)

        if simplify:
            import onnx, onnxsim

            print("Running onnxsim.simplify...")
            tt = onnx.load(filepath)
            tt, check = onnxsim.simplify(tt)
            if check:
                with open(filepath, "wb") as ff:
                    ff.write(tt.SerializeToString())
                print("Exported simplified onnx:", filepath)
            else:
                print("[Error] failed to simplify onnx:", filepath)

    def export_pth(self, filepath=None, input_shape=None, batch_size=1, **kwargs):
        input_datas = self._create_fake_input_data_(input_shape, batch_size=batch_size)

        traced_cell = torch.jit.trace(self, example_inputs=input_datas if len(self.inputs) == 1 else [input_datas])
        filepath = (self.name + ".pth") if filepath is None else (filepath if filepath.endswith(".pth") else (filepath + ".pth"))
        torch.jit.save(traced_cell, filepath, **kwargs)
        print("Exported pth:", filepath)

    def load_weights(self, filepath, by_name=True, skip_mismatch=False, **kwargs):
        if filepath.endswith("h5"):
            from keras_cv_attention_models.download_and_load import load_weights_from_hdf5_file

            load_weights_from_hdf5_file(filepath, self, skip_mismatch=skip_mismatch, debug=self.debug)
        else:
            weights = torch.load(filepath, map_location=torch.device("cpu"), **kwargs)
            weights = weights.state_dict() if hasattr(weights, "state_dict") else weights
            self.load_state_dict(weights.get("state_dict", weights.get("model", weights)))
            if hasattr(self, "optimizer") and "optimizer" in weights:
                print(">>>> Reload optimizer state_dict")
                self.optimizer.load_state_dict(weights["optimizer"])

    def save_weights(self, filepath=None, **kwargs):
        filepath = filepath if filepath else self.name + ".h5"
        if filepath.endswith("h5"):
            from keras_cv_attention_models.download_and_load import save_weights_to_hdf5_file

            save_weights_to_hdf5_file(filepath, self, **kwargs)
        else:
            save_items = {"state_dict": self.state_dict()}
            if hasattr(self, "optimizer"):
                save_items.update({"optimizer": self.optimizer.state_dict()})
            torch.save(save_items, filepath, **kwargs)

    def load(self, filepath, **kwargs):
        self.load_weights(filepath, **kwargs)

    def save(self, filepath=None, **kwargs):
        self.save_weights(filepath, **kwargs)

    def count_params(self):
        total_params = sum([np.prod(ii.shape) for ii in self.state_dict().values() if len(ii.shape) != 0])
        trainable_params = sum([np.prod(list(ii.shape)) for ii in self.parameters()])
        non_trainable_params = total_params - trainable_params
        print("Total params: {:,} | Trainable params: {:,} | Non-trainable params:{:,}".format(total_params, trainable_params, non_trainable_params))
        return total_params


class Model(nn.Module, _Trainer_, _Exporter_):
    """
    Examples:
    # compile and fit on buildin models
    >>> os.environ["KECAM_BACKEND"] = "torch"
    >>> import torch
    >>> from keras_cv_attention_models import aotnet
    >>> mm = aotnet.AotNet50(num_classes=10, input_shape=(32, 32, 3))
    >>> mm.compile(metrics='acc')  # Using default cross_entropy loss
    >>> xx, yy = torch.rand([300, 3, 32, 32]), torch.functional.F.one_hot(torch.randint(0, 10, size=[300]), 10).float()
    >>> mm.fit(xx[:256], yy[:256], epochs=2, validation_data=(xx[256:], yy[256:]))

    # Build custom model
    >>> os.environ["KECAM_BACKEND"] = "torch"
    >>> import torch
    >>> from keras_cv_attention_models.backend import layers, models
    >>> from keras_cv_attention_models.imagenet.callbacks import MyCheckpoint  # Add a callback
    >>> inputs = layers.Input([3, 32, 32])
    >>> nn = layers.Conv2D(32, 3, 2, padding='same')(inputs)
    >>> nn = layers.GlobalAveragePooling2D()(nn)
    >>> nn = layers.Dense(10)(nn)
    >>> mm = models.Model(inputs, nn)
    >>> mm.summary()
    >>> xx, yy = torch.rand([1128, 3, 32, 32]), torch.functional.F.one_hot(torch.randint(0, 10, size=[1128]), 10).float()
    >>> loss = lambda y_true, y_pred: (y_true - y_pred.float()).abs().mean()
    >>> mm.compile(optimizer="AdamW", loss=loss, metrics='acc')
    >>> callbacks = [MyCheckpoint(basic_save_name="test")]
    >>> mm.fit(xx[:1000], yy[:1000], epochs=2, callbacks=callbacks, validation_data=(xx[1000:], yy[1000:]))

    # Prediction using buildin model
    >>> from keras_cv_attention_models.mlp_family import mlp_mixer
    >>> mm = mlp_mixer.MLPMixerB16(input_shape=(3, 224, 224))
    >>> # >>>> Load pretrained from: /home/leondgarse/.keras/models/mlp_mixer_b16_imagenet.h5
    >>> from PIL import Image
    >>> from skimage.data import chelsea # Chelsea the cat
    >>> from keras_cv_attention_models.imagenet import decode_predictions
    >>> imm = Image.fromarray(chelsea()).resize(mm.input_shape[2:])
    >>> pred = mm(torch.from_numpy(np.array(imm)).permute([2, 0, 1])[None] / 255)
    >>> print(decode_predictions(pred.detach()))
    >>> # or just use preset preprocess_input
    >>> print(mm.decode_predictions(mm(mm.preprocess_input(chelsea()))))
    """

    num_instances = 0  # Count instances

    @classmethod
    def __count__(cls):
        cls.num_instances += 1

    def __init__(self, inputs, outputs, name=None, **kwargs):
        super().__init__()
        self.name = "model_{}".format(self.num_instances) if name == None else name
        self.nodes = None

        self.output = outputs
        self.outputs = outputs if isinstance(outputs, (list, tuple)) else [outputs]
        self.output_shape = [tuple(ii.shape) for ii in self.outputs] if isinstance(outputs, (list, tuple)) else tuple(outputs.shape)
        self.output_names = [ii.name for ii in self.outputs]

        self.input = inputs
        self.inputs = inputs if isinstance(inputs, (list, tuple)) else [inputs]
        self.input_shape = [tuple(ii.shape) for ii in self.inputs] if isinstance(inputs, (list, tuple)) else tuple(inputs.shape)
        self.input_names = [ii.name for ii in self.inputs]

        self.num_outputs = len(self.outputs)
        self.create_forward_pipeline()
        self.eval()  # Set eval mode by default
        self.debug = False

    def create_forward_pipeline(self, **kwargs):
        forward_pipeline, all_layers, outputs = [], {}, []
        dfs_queue = []
        intra_nodes_ref = {}  # node names, and how many times they should be used
        for ii in self.inputs:
            dfs_queue.extend(list(set(ii.next_nodes)))
            intra_nodes_ref[ii.name] = len(ii.next_nodes)
        while len(dfs_queue) > 0:
            cur_node = dfs_queue.pop(-1)
            # print(cur_node.name, cur_node.next_nodes)
            if len(cur_node.pre_nodes) > 1 and not all([ii.name in intra_nodes_ref for ii in cur_node.pre_nodes]):
                continue
            if cur_node.name in intra_nodes_ref:
                raise ValueError("All nodes name should be unique: cur_node: {}, intra_nodes_ref: {}".format(cur_node.name, list(intra_nodes_ref.keys())))

            # `set` is used here in case current node outputs multi times to next node, like `all_layers.Add()([inputs, inputs])`.
            # dfs_queue.extend(list(set(cur_node.next_nodes)))
            dfs_queue.extend([ii for ii in cur_node.next_nodes if ii not in dfs_queue])
            # print(f"{dfs_queue = }")

            forward_pipeline.append(cur_node)
            setattr(self, cur_node.name.replace(".", "_"), cur_node.callable)
            all_layers[cur_node.layer.name] = cur_node.layer

            # print(f"{cur_node.name = }, {len(cur_node.next_nodes) = }")
            intra_nodes_ref[cur_node.name] = len(cur_node.next_nodes)
            if cur_node.name in self.output_names:
                intra_nodes_ref[cur_node.name] = intra_nodes_ref.get(cur_node.name, 0) + 1
            if all([ii in intra_nodes_ref for ii in self.output_names]):
                break
        self.forward_pipeline, self.intra_nodes_ref, self.__layers__ = forward_pipeline, intra_nodes_ref, all_layers

    def forward(self, inputs, **kwargs):
        if isinstance(inputs, layers.GraphNode) or (isinstance(inputs, (list, tuple)) and any([isinstance(ii, layers.GraphNode) for ii in inputs])):
            return self.graphnode_forward(inputs)

        # print(' -> '.join([ii.name for ii in self.forward_pipeline]))
        if isinstance(inputs, (list, tuple)):  # Multi inputs in list or tuple format
            intra_nodes = {kk: [vv] * self.intra_nodes_ref[kk] for kk, vv in zip(self.input_names, inputs)}
        elif isinstance(inputs, dict):  # Multi inputs in dict format
            intra_nodes = {kk: [vv] * self.intra_nodes_ref[kk] for kk, vv in inputs.items()}
        else:  # Single input
            intra_nodes = {self.input_names[0]: [inputs] * self.intra_nodes_ref[self.input_names[0]]}

        for node in self.forward_pipeline:
            if self.debug:
                print(">>>> [{}], pre_node_names: {}, next_node_names: {}".format(node.name, node.pre_node_names, node.next_node_names))
                print("     intra_nodes:", {kk: len(vv) for kk, vv in intra_nodes.items() if len(vv) > 0})

            if len(node.pre_nodes) > 1:
                cur_inputs = [intra_nodes[ii].pop() for ii in node.pre_node_names]
            else:
                cur_inputs = intra_nodes[node.pre_node_names[0]].pop()

            if self.debug:
                print("     inputs.shape:", [np.shape(ii) for ii in cur_inputs] if len(node.pre_nodes) > 1 else np.shape(cur_inputs))

            output = node.callable(cur_inputs)
            intra_nodes[node.name] = [output] * self.intra_nodes_ref[node.name]
            if self.debug:
                print("     output.shape:", np.shape(output))
                print("     intra_nodes:", {kk: len(vv) for kk, vv in intra_nodes.items() if len(vv) > 0})
        return [intra_nodes[ii][0] for ii in self.output_names] if self.num_outputs != 1 else intra_nodes[self.output_names[0]][0]

    def graphnode_forward(self, inputs):
        self.input_shape = [() if isinstance(ii, (int, float)) else ii.shape for ii in inputs] if isinstance(inputs, (list, tuple)) else inputs.shape
        cur_node = layers.GraphNode(self.output_shape, name=self.name if self.nodes is None else (self.name + "_{}".format(len(self.nodes))))
        cur_node.callable = self
        cur_node.layer = self
        cur_node.set_pre_nodes(inputs)

        inputs = inputs if isinstance(inputs, (list, tuple)) else [inputs]
        for ii in inputs:
            if isinstance(ii, layers.GraphNode):
                ii.set_next_nodes(cur_node)

        if self.nodes is None:
            self.nodes = [cur_node]
            self.node = cur_node
        else:
            self.nodes.append(cur_node)
        return cur_node

    @property
    def layers(self):
        return list(self.__layers__.values())

    @property
    def weights(self):
        skips = ["num_batches_tracked", "total_ops", "total_params"]
        buffers = [layers.Weight(name=name, value=value) for name, value in self.named_buffers() if not name.split(".")[-1] in skips]
        parameters = [layers.Weight(name=name, value=value) for name, value in self.named_parameters()]
        return parameters + buffers

    def get_layer(self, layer_name):
        return self.__layers__[layer_name]

    def set_debug(self, debug=True):
        self.debug = debug
        print(">>>> debug: {}".format(self.debug))


class Sequential(Model):
    """
    >>> os.environ["KECAM_BACKEND"] = "torch"
    >>> import torch
    >>> from keras_cv_attention_models.pytorch_backend import layers, models, functional
    >>> mm = models.Sequential([
    >>>     layers.Input([3, 32, 32]),
    >>>     layers.Conv2D(32, 3, 2, padding='same'),
    >>>     layers.GlobalAveragePooling2D(),
    >>>     layers.Dense(10),
    >>>     functional.softmax,  # Can also be an functional callable
    >>> ])
    >>> mm.summary()
    >>> print(mm(torch.ones([1, 3, 32, 32])).shape)
    >>> # torch.Size([1, 10])
    """

    def __init__(self, sequence_layers=None, name=None, **kwargs):
        self.sequence_layers, self.name, self.kwargs = sequence_layers, name, kwargs
        if isinstance(sequence_layers[0], layers.Input):
            self.build(sequence_layers[0].shape)
        else:
            self.input_shape, self.built = None, False

    def build(self, input_shape):
        inputs = layers.Input(input_shape[1:])
        next_node = inputs
        for layer in self.sequence_layers:
            if isinstance(layer, layers.Input):
                continue
            next_node = layer(next_node)
        super().__init__(inputs, next_node, name=self.name)
        self.built = True

    def add(self, layer):
        self.sequence_layers.append(layer)
        if self.input_shape is not None:
            self.build(self.input_shape)



"""
Just copied from ultralytics tal.py / Loss function / BboxLoss
"""
import torch
import math
from torch import nn
import torch.nn.functional as F


class Loss:
    def __init__(self, device="cpu", stride=[8, 16, 32], nc=80, reg_max=16, box_weight=7.5, cls_weight=0.5, dfl_weight=1.5, min_memory=False):
        # self.device = next(model.parameters()).device  # get model device
        # h = model.args  # hyperparameters
        # m = model.model[-1]  # Detect() module
        self.bce = nn.BCEWithLogitsLoss(reduction="none")
        self.stride, self.nc, self.reg_max, self.device = stride, nc, reg_max, device
        self.box_weight, self.cls_weight, self.dfl_weight, self.min_memory = box_weight, cls_weight, dfl_weight, min_memory

        self.no = nc + reg_max * 4
        self.use_dfl = reg_max > 1
        roll_out_thr = min_memory if min_memory > 1 else 64 if min_memory else 0  # 64 is default

        self.assigner = TaskAlignedAssigner(topk=10, num_classes=self.nc, alpha=0.5, beta=6.0, roll_out_thr=roll_out_thr)
        self.bbox_loss = BboxLoss(reg_max - 1, use_dfl=self.use_dfl).to(self.device)
        self.proj = torch.arange(reg_max, dtype=torch.float, device=self.device)

    def preprocess(self, targets, batch_size, scale_tensor):
        if targets.shape[0] == 0:
            out = torch.zeros(batch_size, 0, 5, device=self.device)
        else:
            i = targets[:, 0]  # image index
            _, counts = i.unique(return_counts=True)
            out = torch.zeros(batch_size, counts.max(), 5, device=self.device)
            for j in range(batch_size):
                matches = i == j
                n = matches.sum()
                if n:
                    out[j, :n] = targets[matches, 1:]
            out[..., 1:5] = xywh2xyxy(out[..., 1:5].mul_(scale_tensor))
        return out

    def bbox_decode(self, anchor_points, pred_dist):
        if self.use_dfl:
            b, a, c = pred_dist.shape  # batch, anchors, channels
            pred_dist = pred_dist.view(b, a, 4, c // 4).softmax(3).matmul(self.proj.type(pred_dist.dtype))
            # pred_dist = pred_dist.view(b, a, c // 4, 4).transpose(2,3).softmax(3).matmul(self.proj.type(pred_dist.dtype))
            # pred_dist = (pred_dist.view(b, a, c // 4, 4).softmax(2) * self.proj.type(pred_dist.dtype).view(1, 1, -1, 1)).sum(2)
        return dist2bbox(pred_dist, anchor_points, xywh=False)

    def __call__(self, preds, batch):
        loss = torch.zeros(3, device=self.device)  # box, cls, dfl
        feats = preds[1] if isinstance(preds, tuple) else preds
        pred_distri, pred_scores = torch.cat([xi.view(feats[0].shape[0], self.no, -1) for xi in feats], 2).split((self.reg_max * 4, self.nc), 1)

        pred_scores = pred_scores.permute(0, 2, 1).contiguous()
        pred_distri = pred_distri.permute(0, 2, 1).contiguous()

        dtype = pred_scores.dtype
        batch_size = pred_scores.shape[0]
        imgsz = torch.tensor(feats[0].shape[2:], device=self.device, dtype=dtype) * self.stride[0]  # image size (h,w)
        anchor_points, stride_tensor = make_anchors(feats, self.stride, 0.5)

        # targets
        targets = torch.cat((batch["batch_idx"].view(-1, 1), batch["cls"].view(-1, 1), batch["bboxes"]), 1)
        targets = self.preprocess(targets.to(self.device), batch_size, scale_tensor=imgsz[[1, 0, 1, 0]])
        gt_labels, gt_bboxes = targets.split((1, 4), 2)  # cls, xyxy
        mask_gt = gt_bboxes.sum(2, keepdim=True).gt_(0)

        # pboxes
        pred_bboxes = self.bbox_decode(anchor_points, pred_distri)  # xyxy, (b, h*w, 4)

        _, target_bboxes, target_scores, fg_mask, _ = self.assigner(
            pred_scores.detach().sigmoid(),
            (pred_bboxes.detach() * stride_tensor).type(gt_bboxes.dtype),
            anchor_points * stride_tensor,
            gt_labels,
            gt_bboxes,
            mask_gt,
        )

        target_bboxes /= stride_tensor
        target_scores_sum = max(target_scores.sum(), 1)

        # cls loss
        # loss[1] = self.varifocal_loss(pred_scores, target_scores, target_labels) / target_scores_sum  # VFL way
        loss[1] = self.bce(pred_scores, target_scores.to(dtype)).sum() / target_scores_sum  # BCE

        # bbox loss
        if fg_mask.sum():
            loss[0], loss[2] = self.bbox_loss(pred_distri, pred_bboxes, anchor_points, target_bboxes, target_scores, target_scores_sum, fg_mask)

        loss[0] *= self.box_weight  # box gain
        loss[1] *= self.cls_weight  # cls gain
        loss[2] *= self.dfl_weight  # dfl gain

        return loss.sum() * batch_size, loss.detach()  # loss(box, cls, dfl)


def select_candidates_in_gts(xy_centers, gt_bboxes, eps=1e-9, roll_out=False):
    """select the positive anchor center in gt

    Args:
        xy_centers (Tensor): shape(h*w, 4)
        gt_bboxes (Tensor): shape(b, n_boxes, 4)
    Return:
        (Tensor): shape(b, n_boxes, h*w)
    """
    n_anchors = xy_centers.shape[0]
    bs, n_boxes, _ = gt_bboxes.shape
    if roll_out:
        bbox_deltas = torch.empty((bs, n_boxes, n_anchors), device=gt_bboxes.device)
        for b in range(bs):
            lt, rb = gt_bboxes[b].view(-1, 1, 4).chunk(2, 2)  # left-top, right-bottom
            bbox_deltas[b] = torch.cat((xy_centers[None] - lt, rb - xy_centers[None]), dim=2).view(n_boxes, n_anchors, -1).amin(2).gt_(eps)
        return bbox_deltas
    else:
        lt, rb = gt_bboxes.view(-1, 1, 4).chunk(2, 2)  # left-top, right-bottom
        bbox_deltas = torch.cat((xy_centers[None] - lt, rb - xy_centers[None]), dim=2).view(bs, n_boxes, n_anchors, -1)
        # return (bbox_deltas.min(3)[0] > eps).to(gt_bboxes.dtype)
        return bbox_deltas.amin(3).gt_(eps)


def select_highest_overlaps(mask_pos, overlaps, n_max_boxes):
    """if an anchor box is assigned to multiple gts,
        the one with the highest iou will be selected.

    Args:
        mask_pos (Tensor): shape(b, n_max_boxes, h*w)
        overlaps (Tensor): shape(b, n_max_boxes, h*w)
    Return:
        target_gt_idx (Tensor): shape(b, h*w)
        fg_mask (Tensor): shape(b, h*w)
        mask_pos (Tensor): shape(b, n_max_boxes, h*w)
    """
    # (b, n_max_boxes, h*w) -> (b, h*w)
    fg_mask = mask_pos.sum(-2)
    if fg_mask.max() > 1:  # one anchor is assigned to multiple gt_bboxes
        mask_multi_gts = (fg_mask.unsqueeze(1) > 1).repeat([1, n_max_boxes, 1])  # (b, n_max_boxes, h*w)
        max_overlaps_idx = overlaps.argmax(1)  # (b, h*w)
        is_max_overlaps = F.one_hot(max_overlaps_idx, n_max_boxes)  # (b, h*w, n_max_boxes)
        is_max_overlaps = is_max_overlaps.permute(0, 2, 1).to(overlaps.dtype)  # (b, n_max_boxes, h*w)
        mask_pos = torch.where(mask_multi_gts, is_max_overlaps, mask_pos)  # (b, n_max_boxes, h*w)
        fg_mask = mask_pos.sum(-2)
    # find each grid serve which gt(index)
    target_gt_idx = mask_pos.argmax(-2)  # (b, h*w)
    return target_gt_idx, fg_mask, mask_pos


class TaskAlignedAssigner(nn.Module):
    def __init__(self, topk=13, num_classes=80, alpha=1.0, beta=6.0, eps=1e-9, roll_out_thr=0):
        super().__init__()
        self.topk = topk
        self.num_classes = num_classes
        self.bg_idx = num_classes
        self.alpha = alpha
        self.beta = beta
        self.eps = eps
        self.roll_out_thr = roll_out_thr

    @torch.no_grad()
    def forward(self, pd_scores, pd_bboxes, anc_points, gt_labels, gt_bboxes, mask_gt):
        """This code referenced to
           https://github.com/Nioolek/PPYOLOE_pytorch/blob/master/ppyoloe/assigner/tal_assigner.py

        Args:
            pd_scores (Tensor): shape(bs, num_total_anchors, num_classes)
            pd_bboxes (Tensor): shape(bs, num_total_anchors, 4)
            anc_points (Tensor): shape(num_total_anchors, 2)
            gt_labels (Tensor): shape(bs, n_max_boxes, 1)
            gt_bboxes (Tensor): shape(bs, n_max_boxes, 4)
            mask_gt (Tensor): shape(bs, n_max_boxes, 1)
        Returns:
            target_labels (Tensor): shape(bs, num_total_anchors)
            target_bboxes (Tensor): shape(bs, num_total_anchors, 4)
            target_scores (Tensor): shape(bs, num_total_anchors, num_classes)
            fg_mask (Tensor): shape(bs, num_total_anchors)
        """
        self.bs = pd_scores.size(0)
        self.n_max_boxes = gt_bboxes.size(1)
        self.roll_out = self.n_max_boxes > self.roll_out_thr if self.roll_out_thr else False

        if self.n_max_boxes == 0:
            device = gt_bboxes.device
            return (
                torch.full_like(pd_scores[..., 0], self.bg_idx).to(device),
                torch.zeros_like(pd_bboxes).to(device),
                torch.zeros_like(pd_scores).to(device),
                torch.zeros_like(pd_scores[..., 0]).to(device),
                torch.zeros_like(pd_scores[..., 0]).to(device),
            )

        mask_pos, align_metric, overlaps = self.get_pos_mask(pd_scores, pd_bboxes, gt_labels, gt_bboxes, anc_points, mask_gt)

        target_gt_idx, fg_mask, mask_pos = select_highest_overlaps(mask_pos, overlaps, self.n_max_boxes)

        # assigned target
        target_labels, target_bboxes, target_scores = self.get_targets(gt_labels, gt_bboxes, target_gt_idx, fg_mask)

        # normalize
        align_metric *= mask_pos
        pos_align_metrics = align_metric.amax(axis=-1, keepdim=True)  # b, max_num_obj
        pos_overlaps = (overlaps * mask_pos).amax(axis=-1, keepdim=True)  # b, max_num_obj
        norm_align_metric = (align_metric * pos_overlaps / (pos_align_metrics + self.eps)).amax(-2).unsqueeze(-1)
        target_scores = target_scores * norm_align_metric

        return target_labels, target_bboxes, target_scores, fg_mask.bool(), target_gt_idx

    def get_pos_mask(self, pd_scores, pd_bboxes, gt_labels, gt_bboxes, anc_points, mask_gt):
        # get anchor_align metric, (b, max_num_obj, h*w)
        align_metric, overlaps = self.get_box_metrics(pd_scores, pd_bboxes, gt_labels, gt_bboxes)
        # get in_gts mask, (b, max_num_obj, h*w)
        mask_in_gts = select_candidates_in_gts(anc_points, gt_bboxes, roll_out=self.roll_out)
        # get topk_metric mask, (b, max_num_obj, h*w)
        mask_topk = self.select_topk_candidates(align_metric * mask_in_gts, topk_mask=mask_gt.repeat([1, 1, self.topk]).bool())
        # merge all mask to a final mask, (b, max_num_obj, h*w)
        mask_pos = mask_topk * mask_in_gts * mask_gt

        return mask_pos, align_metric, overlaps

    def get_box_metrics(self, pd_scores, pd_bboxes, gt_labels, gt_bboxes):
        if self.roll_out:
            align_metric = torch.empty((self.bs, self.n_max_boxes, pd_scores.shape[1]), device=pd_scores.device)
            overlaps = torch.empty((self.bs, self.n_max_boxes, pd_scores.shape[1]), device=pd_scores.device)
            ind_0 = torch.empty(self.n_max_boxes, dtype=torch.long)
            for b in range(self.bs):
                ind_0[:], ind_2 = b, gt_labels[b].squeeze(-1).long()
                # get the scores of each grid for each gt cls
                bbox_scores = pd_scores[ind_0, :, ind_2]  # b, max_num_obj, h*w
                overlaps[b] = bbox_iou(gt_bboxes[b].unsqueeze(1), pd_bboxes[b].unsqueeze(0), xywh=False, CIoU=True).squeeze(2).clamp(0)
                align_metric[b] = bbox_scores.pow(self.alpha) * overlaps[b].pow(self.beta)
        else:
            ind = torch.zeros([2, self.bs, self.n_max_boxes], dtype=torch.long)  # 2, b, max_num_obj
            ind[0] = torch.arange(end=self.bs).view(-1, 1).repeat(1, self.n_max_boxes)  # b, max_num_obj
            ind[1] = gt_labels.long().squeeze(-1)  # b, max_num_obj
            # get the scores of each grid for each gt cls
            bbox_scores = pd_scores[ind[0], :, ind[1]]  # b, max_num_obj, h*w

            overlaps = bbox_iou(gt_bboxes.unsqueeze(2), pd_bboxes.unsqueeze(1), xywh=False, CIoU=True).squeeze(3).clamp(0)
            align_metric = bbox_scores.pow(self.alpha) * overlaps.pow(self.beta)
        return align_metric, overlaps

    def select_topk_candidates(self, metrics, largest=True, topk_mask=None):
        """
        Args:
            metrics: (b, max_num_obj, h*w).
            topk_mask: (b, max_num_obj, topk) or None
        """

        num_anchors = metrics.shape[-1]  # h*w
        # (b, max_num_obj, topk)
        topk_metrics, topk_idxs = torch.topk(metrics, self.topk, dim=-1, largest=largest)
        if topk_mask is None:
            topk_mask = (topk_metrics.max(-1, keepdim=True) > self.eps).tile([1, 1, self.topk])
        # (b, max_num_obj, topk)
        topk_idxs[~topk_mask] = 0
        # (b, max_num_obj, topk, h*w) -> (b, max_num_obj, h*w)
        if self.roll_out:
            is_in_topk = torch.empty(metrics.shape, dtype=torch.long, device=metrics.device)
            for b in range(len(topk_idxs)):
                is_in_topk[b] = F.one_hot(topk_idxs[b], num_anchors).sum(-2)
        else:
            is_in_topk = F.one_hot(topk_idxs, num_anchors).sum(-2)
        # filter invalid bboxes
        is_in_topk = torch.where(is_in_topk > 1, 0, is_in_topk)
        return is_in_topk.to(metrics.dtype)

    def get_targets(self, gt_labels, gt_bboxes, target_gt_idx, fg_mask):
        """
        Args:
            gt_labels: (b, max_num_obj, 1)
            gt_bboxes: (b, max_num_obj, 4)
            target_gt_idx: (b, h*w)
            fg_mask: (b, h*w)
        """

        # assigned target labels, (b, 1)
        batch_ind = torch.arange(end=self.bs, dtype=torch.int64, device=gt_labels.device)[..., None]
        target_gt_idx = target_gt_idx + batch_ind * self.n_max_boxes  # (b, h*w)
        target_labels = gt_labels.long().flatten()[target_gt_idx]  # (b, h*w)

        # assigned target boxes, (b, max_num_obj, 4) -> (b, h*w)
        target_bboxes = gt_bboxes.view(-1, 4)[target_gt_idx]

        # assigned target scores
        target_labels.clamp(0)
        target_scores = F.one_hot(target_labels, self.num_classes)  # (b, h*w, 80)
        fg_scores_mask = fg_mask[:, :, None].repeat(1, 1, self.num_classes)  # (b, h*w, 80)
        target_scores = torch.where(fg_scores_mask > 0, target_scores, 0)

        return target_labels, target_bboxes, target_scores


def make_anchors(feats, strides, grid_cell_offset=0.5):
    """Generate anchors from features."""
    anchor_points, stride_tensor = [], []
    assert feats is not None
    dtype, device = feats[0].dtype, feats[0].device
    for i, stride in enumerate(strides):
        _, _, h, w = feats[i].shape
        sx = torch.arange(end=w, device=device, dtype=dtype) + grid_cell_offset  # shift x
        sy = torch.arange(end=h, device=device, dtype=dtype) + grid_cell_offset  # shift y
        sy, sx = torch.meshgrid(sy, sx, indexing="ij")
        anchor_points.append(torch.stack((sx, sy), -1).view(-1, 2))
        stride_tensor.append(torch.full((h * w, 1), stride, dtype=dtype, device=device))
    return torch.cat(anchor_points), torch.cat(stride_tensor)


def dist2bbox(distance, anchor_points, xywh=True, dim=-1):
    """Transform distance(ltrb) to box(xywh or xyxy)."""
    lt, rb = distance.chunk(2, dim)
    x1y1 = anchor_points - lt
    x2y2 = anchor_points + rb
    if xywh:
        c_xy = (x1y1 + x2y2) / 2
        wh = x2y2 - x1y1
        return torch.cat((c_xy, wh), dim)  # xywh bbox
    return torch.cat((x1y1, x2y2), dim)  # xyxy bbox


def bbox2dist(anchor_points, bbox, reg_max):
    """Transform bbox(xyxy) to dist(ltrb)."""
    x1y1, x2y2 = bbox.chunk(2, -1)
    return torch.cat((anchor_points - x1y1, x2y2 - anchor_points), -1).clamp(0, reg_max - 0.01)  # dist (lt, rb)


""" BboxLoss """


class BboxLoss(nn.Module):
    def __init__(self, reg_max, use_dfl=False):
        super().__init__()
        self.reg_max = reg_max
        self.use_dfl = use_dfl

    def forward(self, pred_dist, pred_bboxes, anchor_points, target_bboxes, target_scores, target_scores_sum, fg_mask):
        # IoU loss
        weight = torch.masked_select(target_scores.sum(-1), fg_mask).unsqueeze(-1)
        iou = bbox_iou(pred_bboxes[fg_mask], target_bboxes[fg_mask], xywh=False, CIoU=True)
        loss_iou = ((1.0 - iou) * weight).sum() / target_scores_sum

        # DFL loss
        if self.use_dfl:
            target_ltrb = bbox2dist(anchor_points, target_bboxes, self.reg_max)
            loss_dfl = self._df_loss(pred_dist[fg_mask].view(-1, self.reg_max + 1), target_ltrb[fg_mask]) * weight
            loss_dfl = loss_dfl.sum() / target_scores_sum
        else:
            loss_dfl = torch.tensor(0.0).to(pred_dist.device)

        return loss_iou, loss_dfl

    @staticmethod
    def _df_loss(pred_dist, target):
        # Return sum of left and right DFL losses
        # Distribution Focal Loss (DFL) proposed in Generalized Focal Loss https://ieeexplore.ieee.org/document/9792391
        tl = target.long()  # target left
        tr = tl + 1  # target right
        wl = tr - target  # weight left
        wr = 1 - wl  # weight right
        return (
            F.cross_entropy(pred_dist, tl.view(-1), reduction="none").view(tl.shape) * wl
            + F.cross_entropy(pred_dist, tr.view(-1), reduction="none").view(tl.shape) * wr
        ).mean(-1, keepdim=True)


def xywh2xyxy(x):
    """
    Convert bounding box coordinates from (x, y, width, height) format to (x1, y1, x2, y2) format where (x1, y1) is the
    top-left corner and (x2, y2) is the bottom-right corner.

    Args:
        x (np.ndarray) or (torch.Tensor): The input bounding box coordinates in (x, y, width, height) format.
    Returns:
        y (np.ndarray) or (torch.Tensor): The bounding box coordinates in (x1, y1, x2, y2) format.
    """
    y = x.clone() if isinstance(x, torch.Tensor) else np.copy(x)
    y[..., 0] = x[..., 0] - x[..., 2] / 2  # top left x
    y[..., 1] = x[..., 1] - x[..., 3] / 2  # top left y
    y[..., 2] = x[..., 0] + x[..., 2] / 2  # bottom right x
    y[..., 3] = x[..., 1] + x[..., 3] / 2  # bottom right y
    return y


def bbox_iou(box1, box2, xywh=True, GIoU=False, DIoU=False, CIoU=False, eps=1e-7):
    # Returns Intersection over Union (IoU) of box1(1,4) to box2(n,4)

    # Get the coordinates of bounding boxes
    if xywh:  # transform from xywh to xyxy
        (x1, y1, w1, h1), (x2, y2, w2, h2) = box1.chunk(4, -1), box2.chunk(4, -1)
        w1_, h1_, w2_, h2_ = w1 / 2, h1 / 2, w2 / 2, h2 / 2
        b1_x1, b1_x2, b1_y1, b1_y2 = x1 - w1_, x1 + w1_, y1 - h1_, y1 + h1_
        b2_x1, b2_x2, b2_y1, b2_y2 = x2 - w2_, x2 + w2_, y2 - h2_, y2 + h2_
    else:  # x1, y1, x2, y2 = box1
        b1_x1, b1_y1, b1_x2, b1_y2 = box1.chunk(4, -1)
        b2_x1, b2_y1, b2_x2, b2_y2 = box2.chunk(4, -1)
        w1, h1 = b1_x2 - b1_x1, b1_y2 - b1_y1 + eps
        w2, h2 = b2_x2 - b2_x1, b2_y2 - b2_y1 + eps

    # Intersection area
    inter = (b1_x2.minimum(b2_x2) - b1_x1.maximum(b2_x1)).clamp(0) * (b1_y2.minimum(b2_y2) - b1_y1.maximum(b2_y1)).clamp(0)

    # Union Area
    union = w1 * h1 + w2 * h2 - inter + eps

    # IoU
    iou = inter / union
    if CIoU or DIoU or GIoU:
        cw = b1_x2.maximum(b2_x2) - b1_x1.minimum(b2_x1)  # convex (smallest enclosing box) width
        ch = b1_y2.maximum(b2_y2) - b1_y1.minimum(b2_y1)  # convex height
        if CIoU or DIoU:  # Distance or Complete IoU https://arxiv.org/abs/1911.08287v1
            c2 = cw**2 + ch**2 + eps  # convex diagonal squared
            rho2 = ((b2_x1 + b2_x2 - b1_x1 - b1_x2) ** 2 + (b2_y1 + b2_y2 - b1_y1 - b1_y2) ** 2) / 4  # center dist ** 2
            if CIoU:  # https://github.com/Zzh-tju/DIoU-SSD-pytorch/blob/master/utils/box/box_utils.py#L47
                v = (4 / math.pi**2) * (torch.atan(w2 / h2) - torch.atan(w1 / h1)).pow(2)
                with torch.no_grad():
                    alpha = v / (v - iou + (1 + eps))
                return iou - (rho2 / c2 + v * alpha)  # CIoU
            return iou - rho2 / c2  # DIoU
        c_area = cw * ch + eps  # convex area
        return iou - (c_area - union) / c_area  # GIoU https://arxiv.org/pdf/1902.09630.pdf
    return iou  # IoU



import os
import numpy as np


def walk_data_path_gather_images(data_path):
    all_images = []
    for cur, dirs, files in os.walk(data_path):
        all_images.extend([os.path.join(cur, ii) for ii in files if os.path.splitext(ii)[-1].lower() in [".jpg", ".png"]])
    return all_images


def init_from_json(data_path):
    import json

    with open(data_path, "r") as ff:
        aa = json.load(ff)
    train, info = aa["train"], aa.get("info", {})

    all_images = [ii["image"] for ii in train]
    all_labels = [ii["label"] for ii in train] if "label" in train[0] else []
    total_images, num_classes = len(train), info.get("num_classes", 0)

    if num_classes <= 0 and "label" in train[0] and isinstance(train[0]["label"], int):
        num_classes = max(all_labels) + 1
        print(">>>> Using max value from train as num_classes:", num_classes)
    if "base_path" in info and len(info["base_path"]) > 0:
        base_path = os.path.expanduser(info["base_path"])
        all_images = [os.path.join(base_path, ii) for ii in all_images]
    return all_images, all_labels, num_classes


def init_diffusion_alpha(num_training_steps=1000, beta_max=0.02):
    beta = np.linspace(0.0001, beta_max, num_training_steps).astype("float32")[:, None, None, None]  # expand to calculation on batch dimension
    alpha_bar = np.cumprod(1.0 - beta, axis=0)
    sqrt_alpha_bar = alpha_bar**0.5
    sqrt_one_minus_alpha_bar = (1 - alpha_bar) ** 0.5
    return sqrt_alpha_bar, sqrt_one_minus_alpha_bar


def build_torch_dataset(images, labels=None, image_size=512, batch_size=32, num_training_steps=1000, use_horizontal_flip=True):
    import torch
    from PIL import Image
    from torch.utils.data import DataLoader, Dataset
    from torchvision.transforms import Normalize, Compose, RandomHorizontalFlip, Resize, InterpolationMode, ToTensor

    use_labels = False if labels is None or len(labels) == 0 else True
    image_size = image_size if isinstance(image_size, (list, tuple)) else (image_size, image_size)
    sqrt_alpha_bar, sqrt_one_minus_alpha_bar = init_diffusion_alpha(num_training_steps)
    sqrt_alpha_bar, sqrt_one_minus_alpha_bar = torch.from_numpy(sqrt_alpha_bar), torch.from_numpy(sqrt_one_minus_alpha_bar)

    if use_labels:
        labels = [ii + 1 for ii in labels]  # add 1 to labels for skipping 0 as non-conditional
        # to_zero_labels_rate = 1 / max(labels)
        # print(">>>> dataset to_zero_labels_rate:", to_zero_labels_rate)

    class _Dataset_(Dataset):
        def __init__(self, images, labels=None, image_size=512):
            self.images, self.labels = images, labels
            self.mean, self.std = (0.5, 0.5, 0.5), (0.5, 0.5, 0.5)
            self.transforms = Compose(
                [
                    Resize(image_size, interpolation=InterpolationMode.BICUBIC),
                    RandomHorizontalFlip() if use_horizontal_flip else lambda image: image,
                    lambda image: image.convert("RGB"),
                    ToTensor(),
                    Normalize(mean=self.mean, std=self.std),
                ]
            )

        def __len__(self):
            return len(self.images)

        def __getitem__(self, idx):
            image = self.transforms(Image.open(str(self.images[idx])))
            return (image, self.labels[idx]) if use_labels else image

    def diffusion_process(batch):
        if use_labels:
            images, labels = list(zip(*batch))
            images = torch.stack(images)
            labels = torch.tensor(labels)  # if torch.rand(()) > to_zero_labels_rate else torch.zeros([batch_size], dtype=torch.int64)  # [perform worse ???]
        else:
            images = torch.stack(batch)
        timestep = torch.randint(num_training_steps, size=(batch_size,))
        noise = torch.randn_like(images)

        xt = sqrt_alpha_bar[timestep] * images + sqrt_one_minus_alpha_bar[timestep] * noise
        return ((xt, labels, timestep), noise) if use_labels else ((xt, timestep), noise)

    dd = _Dataset_(images, labels, image_size)
    return DataLoader(dd, batch_size=batch_size, collate_fn=diffusion_process, shuffle=True, num_workers=4, drop_last=True, pin_memory=True)


def build_tf_dataset(images, labels=None, image_size=512, batch_size=32, num_training_steps=1000, use_horizontal_flip=True):
    import tensorflow as tf
    from keras_cv_attention_models.imagenet.data import tf_imread

    use_labels = False if labels is None or len(labels) == 0 else True
    image_size = image_size if isinstance(image_size, (list, tuple)) else (image_size, image_size)
    sqrt_alpha_bar, sqrt_one_minus_alpha_bar = init_diffusion_alpha(num_training_steps)
    sqrt_alpha_bar, sqrt_one_minus_alpha_bar = tf.convert_to_tensor(sqrt_alpha_bar), tf.convert_to_tensor(sqrt_one_minus_alpha_bar)
    AUTOTUNE, buffer_size, seed = tf.data.AUTOTUNE, batch_size * 100, None

    if use_labels:
        labels = [ii + 1 for ii in labels]  # add 1 to labels for skipping 0 as non-conditional
        train_dataset = tf.data.Dataset.from_tensor_slices((images, labels)).shuffle(buffer_size=buffer_size, seed=seed)
        # to_zero_labels_rate = 1 / max(labels)
        # print(">>>> dataset to_zero_labels_rate:", to_zero_labels_rate)
    else:
        train_dataset = tf.data.Dataset.from_tensor_slices(images).shuffle(buffer_size=buffer_size, seed=seed)
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    train_dataset = train_dataset.apply(tf.data.experimental.assert_cardinality(len(images))).with_options(options)

    def image_process(image, label=None):
        image = tf_imread(image)
        image = tf.image.resize(image, image_size, method="bicubic", antialias=True)
        image = tf.image.random_flip_left_right(image) if use_horizontal_flip else image
        image = tf.cast(image, tf.float32)
        image.set_shape([*image_size, 3])
        return (image, label) if use_labels else image

    def diffusion_process(image, label=None):
        timestep = tf.random.uniform([batch_size], 0, num_training_steps, dtype="int64")
        noise = tf.random.normal([batch_size, *image_size, 3])
        image = image / 127.5 - 1
        # labels = tf.cond(tf.random.uniform(()) > to_zero_labels_rate, lambda: label, lambda: tf.zeros([batch_size], dtype=label.dtype))

        xt = tf.gather(sqrt_alpha_bar, timestep) * image + tf.gather(sqrt_one_minus_alpha_bar, timestep) * noise
        return ((xt, label, timestep), noise) if use_labels else ((xt, timestep), noise)

    train_dataset = train_dataset.map(image_process, num_parallel_calls=AUTOTUNE)
    train_dataset = train_dataset.batch(batch_size, drop_remainder=True).map(diffusion_process, num_parallel_calls=AUTOTUNE)
    train_dataset = train_dataset.prefetch(buffer_size=AUTOTUNE)
    return train_dataset



import os
import inspect
import torch
import hashlib

_GLOBAL_CUSTOM_OBJECTS = {}
_GLOBAL_CUSTOM_NAMES = {}


def register_keras_serializable(package="Custom", name=None):
    def decorator(arg):
        """Registers a class with the Keras serialization framework."""
        class_name = name if name is not None else arg.__name__
        registered_name = package + ">" + class_name

        if inspect.isclass(arg) and not hasattr(arg, "get_config"):
            raise ValueError("Cannot register a class that does not have a " "get_config() method.")

        # if registered_name in _GLOBAL_CUSTOM_OBJECTS:
        #     raise ValueError(f"{registered_name} has already been registered to " f"{_GLOBAL_CUSTOM_OBJECTS[registered_name]}")

        # if arg in _GLOBAL_CUSTOM_NAMES:
        #     raise ValueError(f"{arg} has already been registered to " f"{_GLOBAL_CUSTOM_NAMES[arg]}")
        _GLOBAL_CUSTOM_OBJECTS[registered_name] = arg
        _GLOBAL_CUSTOM_NAMES[arg] = registered_name

        return arg

    return decorator


def validate_file_md5(fpath, file_hash, chunk_size=65535):
    """Validates a file against a md5 hash. From keras/utils/data_utils.py"""
    hasher = hashlib.md5()
    with open(fpath, "rb") as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b""):
            hasher.update(chunk)
    return str(hasher.hexdigest()) == str(file_hash)


def _extract_archive(file_path, path=".", archive_format="auto"):
    if "zip" in archive_format or (archive_format == "auto" and file_path.endswith(".zip")):
        import zipfile

        assert zipfile.is_zipfile(file_path), "Not a zip file: {}".format(file_path)
        open_fn = zipfile.ZipFile
    elif "tar" in archive_format or (archive_format == "auto" and (file_path.endswith(".tar") or file_path.endswith(".tar.gz"))):
        import tarfile

        assert tarfile.is_tarfile(file_path), "Not a tar file: {}".format(file_path)
        open_fn = tarfile.open
    else:
        raise ValueError("Not a supported extract file format: {}".format(file_path))

    print(">>>> Extract {} -> {}".format(file_path, path))
    with open_fn(file_path) as ff:
        ff.extractall(path)
    return path


def get_file(fname=None, origin=None, cache_subdir="datasets", file_hash=None, extract=False):
    # print(f">>>> {fname = }, {origin = }, {cache_subdir = }, {file_hash = }")
    save_dir = os.path.join(os.path.expanduser("~"), ".keras", cache_subdir)
    if not os.path.exists(save_dir):
        os.makedirs(save_dir, exist_ok=True)
    fname = os.path.basename(origin) if fname is None else fname
    file_path = os.path.join(save_dir, fname)
    if os.path.exists(file_path):
        if file_hash is not None and not validate_file_md5(file_path, file_hash):
            print(
                "A local file was found, but it seems to be incomplete or outdated because the md5 file hash does not match the original value of "
                f"{file_hash} so we will re-download the data."
            )
        else:
            return file_path

    print("Downloading data from {} to {}".format(origin, file_path))
    torch.hub.download_url_to_file(origin, file_path)
    if os.path.exists(file_path) and file_hash is not None and not validate_file_md5(file_path, file_hash):
        raise ValueError("Incomplete or corrupted file detected. The md5 file hash does not match the provided value {}.".format(file_hash))

    if extract:
        _extract_archive(file_path, path=save_dir)  # return tar file path, just like keras one
    return file_path
